#define GTK218
//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include <unistd.h>
#include <stdio.h>
#ifdef _WIN32
    #include <windows.h>
    #include <gdk/gdkwin32.h>
    #include <fcntl.h>
	#define pipe(phandles)	_pipe (phandles, 4096, _O_BINARY)
#else
    #define Sleep(x) usleep(x*1000)
#endif

#include <stdarg.h>

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include "AnsiString.h"
#include <SDL_getenv.h>
#include <list>

#ifndef _WIN32
    #include <gdk/gdkx.h>
#endif

extern "C" {
    #include "ffplay.h"
}


#define CUSTOM_CONTROL_ID       1002

#define MEDIA_ENABLE_AUDIO      10003

#define MEDIA_OPEN_FILE         10006
#define MEDIA_OPEN_STREAM       10007
#define MEDIA_CLOSE             10008

#define MEDIA_PLAY              10009
#define MEDIA_PAUSE             10010
#define MEDIA_STOP              10011

#define MEDIA_STEP              10012
#define MEDIA_SEEK              10014
#define MEDIA_SKIP              10015

#define MEDIA_VIDEO_WIDTH       10015
#define MEDIA_VIDEO_HEIGHT      10016

#define MEDIA_HAS_VIDEO         10017
#define MEDIA_HAS_AUDIO         10018
#define MEDIA_HAS_SUBTITLE      10019

#define MEDIA_INFO_TITLE        10023
#define MEDIA_INFO_AUTHOR       10024
#define MEDIA_INFO_COPYRIGHT    10025
#define MEDIA_INFO_COMMENT      10026
#define MEDIA_INFO_ALBUM        10027
#define MEDIA_INFO_YEAR         10028
#define MEDIA_INFO_TRACK        10029
#define MEDIA_INFO_GENRE        10030
#define MEDIA_INFO_OFFSET       10031
#define MEDIA_INFO_DURATION     10032
#define MEDIA_INFO_SIZE         10033
#define MEDIA_INFO_STARTTIME    10034
#define MEDIA_CURRENT_FPS       10035
#define MEDIA_CURRENT_TIME      10036
#define MEDIA_CODEC_NAME        10037
#define MEDIA_INFO_PACKETSIZE   10038
#define MEDIA_BUFFERS_COUNT     10039
#define MEDIA_CURSOR            10040





INVOKER RemoteInvoker=0;
static AnsiString res_data;
static int pos=0;
static AnsiString stream_buffer;
VideoState *stream=0;
int my_pipe[2];
int pipe_set=0;
AnsiString buffered_text;
int controls_count=0;

bool is_init=false;
bool events_init=false;
long rec_bytes=0;


#define CHECK_STREAM(msg) if (!stream) { ShowError("Cannot stop: No stream opened"); break; }


void ShowError(char *text) {
    GtkWidget *dialog=gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "%s", "Error in media control");
    gtk_window_set_title(GTK_WINDOW(dialog), "Error");
    gtk_window_set_keep_above(GTK_WINDOW(dialog), FALSE);
    gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
    gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog), "%s", text);

    gtk_dialog_run(GTK_DIALOG(dialog));
}

#define EVENT_PLAY_VIDEORESIZE    350
#define EVENT_PLAY_KEYPRESS       351
#define EVENT_PLAY_MOUSE          352
#define EVENT_PLAY_FINISHED       353
#define EVENT_PLAY_ERROR          354
#define EVENT_PLAY_ERROR_FATAL    355

char note_event[6]={0,0,0,0,0,0};
std::list<AnsiString *> buffer_list;
SDL_mutex *m_mutex=0;
//SDL_cond *m_cond=0;
SDL_Thread *buf_tid=0;
void *owner_context=0;


void clear_buffer() {
    SDL_LockMutex(m_mutex);
    std::list<AnsiString *>::const_iterator it=buffer_list.begin();
    for(it=buffer_list.begin(); it!=buffer_list.end(); ++it) {
            AnsiString *data=*it;

            if (data)
                delete data;
    }
    if (!buffer_list.empty())
        buffer_list.clear();
    SDL_UnlockMutex(m_mutex);
}


int BufferCount() {
    SDL_LockMutex(m_mutex);
    int count=buffer_list.size();
    SDL_UnlockMutex(m_mutex);
    return count;
    //return rec_bytes;
}

int WriteThread(void *ptr) {
    while (true) {
        SDL_LockMutex(m_mutex);

        if ((pipe_set) && (buffer_list.size())) {
            std::list<AnsiString *>::const_iterator it=buffer_list.begin();

            AnsiString *data=*it;
            buffer_list.pop_front();

            SDL_UnlockMutex(m_mutex);

            if (data) {
                int len=data->Length();
                if (len) {
                    write(my_pipe[1], data->c_str(), len);
                    rec_bytes-=len;
                }
                delete data;
            }
        } else {
            SDL_UnlockMutex(m_mutex);
            SDL_Delay(10);
        }
    }
    return 0;
}

int play_event(int event, void *data) {
    switch ((PlayEvents)event) {
        case PLAY_ERROR:
            if (note_event[PLAY_ERROR])
                RemoteInvoker(owner_context, INVOKE_FIREEVENT, (int)EVENT_PLAY_ERROR, (char *)data);
            //ShowError((char *)data);
            //std::cerr << "ERROR:" << (char *)data;
            break;
        case PLAY_ERROR_FATAL:
            if (note_event[PLAY_ERROR_FATAL])
                RemoteInvoker(owner_context, INVOKE_FIREEVENT, (int)EVENT_PLAY_ERROR_FATAL, (char *)data);

            //ShowError((char *)data);
            //std::cerr << "ERROR:" << (char *)data;
            break;
        case PLAY_FINISHED:
            if (note_event[PLAY_FINISHED])
                RemoteInvoker(owner_context, INVOKE_FIREEVENT, (int)EVENT_PLAY_FINISHED, (char *)"0");
            if (stream) {
                stream_close(stream);
                stream=stream_new(stream->filename);
            }
            break;
        case PLAY_VIDEORESIZE:
            if (note_event[PLAY_VIDEORESIZE])
                RemoteInvoker(owner_context, INVOKE_FIREEVENT, (int)EVENT_PLAY_VIDEORESIZE, (char *)"0");
            break;
        case PLAY_KEYPRESS:
            if (note_event[PLAY_KEYPRESS]) {
                RemoteInvoker(owner_context, INVOKE_FIREEVENT, (int)EVENT_PLAY_KEYPRESS, AnsiString((long)(int)data).c_str());
            }
            break;
        case PLAY_MOUSE:
            if (note_event[PLAY_MOUSE])
                RemoteInvoker(owner_context, INVOKE_FIREEVENT, (int)EVENT_PLAY_MOUSE, AnsiString((long)(int)data).c_str());
            break;
        default:
            break;
    }
    return 0;
}
// Control invoker function
void init(GtkWidget *widget) {
#ifdef GTK218
    gdk_window_ensure_native(widget->window);
#endif
#ifdef _WIN32
    if ((!is_init) && (widget) && (widget->window) && (GDK_WINDOW_HWND(widget->window))) {
#else
    if ((!is_init) && (widget) && (widget->window) && (GDK_WINDOW_XWINDOW(widget->window))) {
#endif
        is_init=true;

		long wid=0;
#ifdef _WIN32
		wid=(long)GDK_WINDOW_HWND(widget->window);
	    char SDL_windowhack[256];
	    sprintf(SDL_windowhack,"SDL_WINDOWID=%ld",GDK_WINDOW_HWND(widget->window));
	    SDL_putenv(SDL_windowhack);

        /*if (widget->parent)
            MoveWindow(hWnd,widget->parent->allocation.x, widget->parent->allocation.y, widget->parent->allocation.width, widget->parent->allocation.height, TRUE);*/
#else
		wid=(long)GDK_WINDOW_XWINDOW(widget->window);
	    char SDL_windowhack[256];
	    sprintf(SDL_windowhack,"%ld",GDK_WINDOW_XWINDOW(widget->window));
	    //setenv("SDL_WINDOWID", SDL_windowhack, 1);
		SDL_putenv(SDL_windowhack);
#endif
        init_player(play_event, wid);
        events_init=true;
    }
}


#ifdef _WIN32 
    #ifdef GTK218
        static gboolean on_sdl_expose(GtkWidget *viewport, GdkEventExpose *event, gpointer userdata)
        {
            HWND wnd=(HWND)GDK_WINDOW_HWND(viewport->window);

            SendMessage(wnd,WM_PAINT,0,0);
            return FALSE;
        }

        static gboolean on_sdl_resize(GtkWidget *viewport, GdkRectangle *allocation, gpointer userdata)
        {
            HWND wnd=(HWND)GDK_WINDOW_HWND(viewport->window);
            int x=allocation->x;
            int y=allocation->y;
            GtkWidget *parent=viewport->parent;
            AnsiString S=" ";
            if (parent) {
                GdkWindowType wt=gdk_window_get_window_type(parent->window);
                if (wt == GDK_WINDOW_CHILD) {
                    S+=" ";
                    if (parent->allocation.x>0)
                        x+=parent->allocation.x;
                    if (parent->allocation.y>0)
                        y+=parent->allocation.y;
                    int border=gtk_container_get_border_width(GTK_CONTAINER(parent));
                    x+=border;
                    y+=border;
                }
                //parent=parent->parent;
            }
            MoveWindow(wnd,x,y,allocation->width,allocation->height,TRUE);
            SendMessage(wnd,WM_PAINT,0,0);
            return FALSE;
        }
    #endif
#endif


PREFIX int Invoke(void *context,int INVOKE_TYPE,...) {
    va_list arg;

    switch (INVOKE_TYPE) {
        case INVOKE_QUERYTEST:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id=va_arg(arg, int);
                va_end(arg);

                if (cls_id==CUSTOM_CONTROL_ID)
                    return E_ACCEPTED;
            }
            break;
        case INVOKE_SETEVENT:
            {
                if (!RemoteInvoker)
                    break;
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int event_id    =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                if (!owner_context)
                    owner_context=context;

                //init(((Gtk::Widget *)context)->gobj());
                if (!events_init) {
                    events_init=true;
                    register_events(play_event);
                }

                switch (event_id) {
                    case EVENT_PLAY_VIDEORESIZE:
                            note_event[PLAY_VIDEORESIZE]=1;
                            return E_ACCEPTED;
                    case EVENT_PLAY_KEYPRESS:
                            note_event[PLAY_KEYPRESS]=1;
                            return E_ACCEPTED;
                    case EVENT_PLAY_MOUSE:
                            note_event[PLAY_MOUSE]=1;
                            return E_ACCEPTED;
                    case EVENT_PLAY_FINISHED:
                            note_event[PLAY_FINISHED]=1;
                            return E_ACCEPTED;
                    case EVENT_PLAY_ERROR:
                            note_event[PLAY_ERROR]=1;
                            return E_ACCEPTED;
                    case EVENT_PLAY_ERROR_FATAL:
                            note_event[PLAY_ERROR_FATAL]=1;
                            return E_ACCEPTED;
                    default:
                        return E_NOTIMPLEMENTED;
                }
                break;
            }
            break;
        case INVOKE_SETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID)) {
                    break;
                }
                init(GTK_WIDGET(context));

	            //HtmlView *view=HTML_VIEW(((Gtk::Widget *)context)->gobj());
                AnsiString data;
                data.LoadBuffer(str_buffer,buf_len);

                switch (prop_id) {
                        case MEDIA_PLAY:
                            if (!data.ToInt())
                                break;
                            if (!stream) {
                                ShowError("Cannot Play: No stream opened"); 
                                break;
                            }
                            if (stream->paused)
                                toggle_pause();
                            stream_seek(stream, 0, 0);
                            break;
                        case MEDIA_OPEN_FILE:
                            if ((stream) && (!stream->paused))
                                toggle_pause();
                            SDL_LockMutex(m_mutex);
                            clear_buffer();
                            pipe_set=0;
                            SDL_UnlockMutex(m_mutex);

                            stream=stream_new(str_buffer);
                            if (stream){
                                resize(GTK_WIDGET(context)->allocation.width, GTK_WIDGET(context)->allocation.height);
                            }
                            break;
                        case MEDIA_OPEN_STREAM:
                            {
                                if ((stream) && (!stream->paused)) {
                                    toggle_pause();
    							    Sleep(1000);
                                }
                                AnsiString pipe_str("pipe:");
                                if (!data.ToInt())
                                    break;
                                rec_bytes=0;
                                if (pipe_set) {
                                    close(my_pipe[0]);
                                    close(my_pipe[1]);
                                }
                                pipe(my_pipe);
                                pipe_set=1;

                                SDL_LockMutex(m_mutex);
                                clear_buffer();
                                SDL_UnlockMutex(m_mutex);

                                pipe_str+=AnsiString((long)my_pipe[0]);
                                stream=stream_new(pipe_str);
                                if (stream)
                                    resize(GTK_WIDGET(context)->allocation.width, GTK_WIDGET(context)->allocation.height);
    							Sleep(1000);
                            }
                            break;
                        case MEDIA_CLOSE:
                            {
                                int old_pipe_set=pipe_set;
                                if (!data.ToInt())
                                    break;
                                if (!stream) {
                                    ShowError("Cannot Close: No stream opened"); 
                                    break;
                                }

                                if (pipe_set) {
                                    SDL_LockMutex(m_mutex);
                                    SDL_KillThread(buf_tid);
                                    clear_buffer();
                                    pipe_set=0;
                                    SDL_UnlockMutex(m_mutex);
                                    buf_tid = SDL_CreateThread(WriteThread,0);
                                }
                                if (stream) {
                                    stream_close(stream);
                                }
                                stream=0;
                                if (old_pipe_set) {
                                    /*int len=0;
                                    do {
                                        len=read(my_pipe[0], buffer, 4096);
                                    } while (len!=4096);*/
                                    //close(my_pipe[1]);
                                    //close(my_pipe[0]);
                                }
    							Sleep(500);
                            }
                            break;
                        case MEDIA_PAUSE:
                            if (!stream) {
                                ShowError("Cannot Pause: No stream opened"); 
                                break;
                            }
                            if (data.ToInt()) {
                                if (!stream->paused)
                                    toggle_pause();
                            } else {
                                if (stream->paused)
                                    toggle_pause();
                            }
                            break;
                        case MEDIA_STEP:
                            if (!data.ToInt())
                                break;
                            if (!stream) {
                                ShowError("Cannot Step: No stream opened"); 
                                break;
                            }
                            step_to_next_frame();
                            break;
                        case MEDIA_ENABLE_AUDIO:
                            if (!stream) {
                                ShowError("Cannot enable/disable audio : No stream opened"); 
                                break;
                            }
                            if (data.ToInt()) {
                                if (!stream->show_audio)
                                    toggle_audio_display();
                            } else {
                                if (stream->show_audio)
                                    toggle_audio_display();
                            }
                            break;
                        case MEDIA_SEEK:
                            if (stream) {
                                double incr=data.ToFloat(), pos;
                                pos = url_ftell(stream->ic->pb);
                                if (stream->ic->bit_rate)
                                    incr *= stream->ic->bit_rate / 60.0;
                                else
                                    incr *= 180000.0;
                                pos += incr;
                                stream_seek(stream, pos, incr);                              
                            } else {
                                ShowError("Cannot seek : No stream opened"); 
                                break;
                            }
                            break;
                        case MEDIA_SKIP:
                            if (stream) {
                                double incr=data.ToFloat(), pos;
                                pos = get_master_clock(stream);
                                pos += incr;
                                stream_seek(stream, (int64_t)(pos * AV_TIME_BASE), incr);
                            } else {
                                ShowError("Cannot seek : No stream opened"); 
                                break;
                            }
                            break;
                        case MEDIA_STOP:
                            if (!stream) {
                                ShowError("Cannot stop: No stream opened"); 
                                break;
                            }
                            pos = url_ftell(stream->ic->pb);
                            stream_seek(stream, pos, -pos); 
							Sleep(500);

                            if (!stream->paused)
                                toggle_pause();
                            break;
                        case MEDIA_CURSOR:
                            SDL_ShowCursor(data.ToInt());
                            break;
                }
                
                return E_NOERROR;
            }
            break;
        case INVOKE_GETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char **str_buffer=   va_arg(arg, char **);
                int *buf_len     =   va_arg(arg, int*);
                va_end(arg);

                //MPEG *mpg=0;
                //RemoteInvoker(context,INVOKE_GETEXTRAPTR,&mpg);
                //mpg = (MPEG*) gtk_object_get_data( GTK_OBJECT( ((Gtk::Widget *)context)->gobj() ), "mpeg" );

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;


                switch (prop_id) {
                    case MEDIA_VIDEO_WIDTH:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video width: No open stream");
                        res_data=AnsiString((long)stream->width);
                        break;
                    case MEDIA_VIDEO_HEIGHT:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video height: No open stream");
                        res_data=AnsiString((long)stream->height);
                        break;
                    case MEDIA_INFO_TITLE:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video information: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((char *)stream->ic->title);
                        break;
                    case MEDIA_INFO_AUTHOR:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video information: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((char *)stream->ic->author);
                        break;
                    case MEDIA_INFO_COPYRIGHT:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video information: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((char *)stream->ic->copyright);
                        break;
                    case MEDIA_INFO_COMMENT:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video information: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((char *)stream->ic->comment);
                        break;
                    case MEDIA_INFO_ALBUM:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video information: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((char *)stream->ic->album);
                        break;
                    case MEDIA_INFO_YEAR:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video information: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((long)stream->ic->year);
                        break;
                    case MEDIA_INFO_TRACK:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video information: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((long)stream->ic->track);
                        break;
                    case MEDIA_INFO_GENRE:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video info: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((char *)stream->ic->genre);
                        break;
                    case MEDIA_INFO_OFFSET:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video info: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((long)stream->ic->data_offset);
                        break;
                    case MEDIA_INFO_PACKETSIZE:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video info: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((long)stream->ic->packet_size);
                        break;
                    case MEDIA_INFO_DURATION:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video info: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((long)stream->ic->duration);
                        break;
                    case MEDIA_INFO_SIZE:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video info: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((long)stream->ic->file_size);
                        break;
                    case MEDIA_INFO_STARTTIME:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video info: No open stream");
                        if (!stream->ic)
                            break;
                        res_data=AnsiString((long)stream->ic->start_time);
                        break;
                    case MEDIA_CURRENT_FPS:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video info: No open stream");
                        if ((!stream->ic) || (!stream->ic->cur_st))
                            break;
                        res_data=AnsiString((double)stream->ic->cur_st->r_frame_rate.num/(double)stream->ic->cur_st->r_frame_rate.den);
                        break;
                    case MEDIA_CURRENT_TIME:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video info: No open stream");
                        if ((!stream->ic) || (!stream->ic->cur_st))
                            break;
                        res_data=AnsiString((long)stream->ic->cur_st->start_time);
                        break;
                    case MEDIA_CODEC_NAME:
                        res_data="-1";
                        CHECK_STREAM("Cannot return video info: No open stream");
                        if ((!stream->ic) || (!stream->ic->cur_st) || (!stream->ic->cur_st->codec))
                            break;
                        res_data=AnsiString((char *)stream->ic->cur_st->codec->codec_name);
                        break;
                    case MEDIA_HAS_VIDEO:
                        res_data="0";
                        CHECK_STREAM("Cannot return video height: No open stream");
                        if (stream->video_st)
                            res_data="1";
                        break;
                    case MEDIA_HAS_AUDIO:
                        res_data="0";
                        CHECK_STREAM("Cannot return video height: No open stream");
                        if (stream->audio_st)
                            res_data="1";
                        break;
                    case MEDIA_HAS_SUBTITLE:
                        res_data="0";
                        CHECK_STREAM("Cannot return video height: No open stream");
                        if (stream->subtitle_st)
                            res_data="1";
                        break;
                    case MEDIA_BUFFERS_COUNT:
                        res_data=AnsiString((long)BufferCount());
                        break;
                    default:
                        return E_NOTIMPLEMENTED;
                }
                *str_buffer=res_data.c_str();
                *buf_len=res_data.Length();                
                return E_NOERROR;
            }
            break;
        case INVOKE_SETSTREAM:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                char *target    =   va_arg(arg, char *);
                int len_target  =   va_arg(arg, int);
                char *value     =   va_arg(arg, char *);
                int len_value   =   va_arg(arg, int);
                va_end(arg);


                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID)) {
                    break;
                }

                if (pipe_set) {
                    SDL_LockMutex(m_mutex);
                    //write(my_pipe[1], value, len_value);
                    rec_bytes+=len_value;
                    AnsiString *buf=new AnsiString();
                    buf->LoadBuffer(value, len_value);
                    buffer_list.push_back(buf);
                    SDL_UnlockMutex(m_mutex);
                } else {
                    ShowError("The stream you're using is unitialized");
                }

                return E_NOERROR;
            }
            break;
        case INVOKE_SETSECONDARYSTREAM:
            // not needed
            break;
        case INVOKE_CUSTOMMESSAGE:
            {
                va_start(arg, INVOKE_TYPE);
                int     msg     =   va_arg(arg, int);
                int     cls_id  =   va_arg(arg, int);
                char   *target  =   va_arg(arg, char*);
                int     t_len   =   va_arg(arg, int);
                char   *value   =   va_arg(arg, char*);
                int     v_len   =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID)) {
                    break;
                }

                /*AnsiString a_target(target);
                AnsiString a_value(value);

                switch (msg) {
                    case MSG_CUSTOM_MESSAGE1:
                        Mix_Volume(a_target.ToInt(), a_value.ToInt());
                        return E_ACCEPTED;
                }*/
            }
            break;
        case INVOKE_CREATE:
            {
                if (controls_count>0) {
                    ShowError("For now, only one multimedia control is supported per application");
                    return E_NOTIMPLEMENTED;
                }
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void **handler  =   va_arg(arg, void **);
                int *pack_type  =   va_arg(arg, int *);
                va_end(arg);
                *pack_type=(int)0;

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;


                GtkWidget *wrapped=gtk_drawing_area_new();
#ifdef GTK218
                gdk_window_ensure_native(wrapped->window);
                g_signal_connect(
                           G_OBJECT(wrapped), 
                           "size-allocate", 
                           G_CALLBACK(on_sdl_resize), 
                           0
                );

                g_signal_connect(
                           G_OBJECT(wrapped), 
                           "expose_event", 
                           G_CALLBACK(on_sdl_expose), 
                           0
                );
                /*gtk_widget_set_app_paintable(GTK_WIDGET(wrapped),FALSE);*/

#endif

                //GtkWidget *widget=wrapped->gobj();

                controls_count++;

                *handler=wrapped;
                return E_ACCEPTED;
            }
            break;
        case INVOKE_DESTROY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void *handler  =   va_arg(arg, void *);
                va_end(arg);

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                controls_count--;

                owner_context=0;

                return E_ACCEPTED;
            }
            break;
        case INVOKE_INITLIBRARY:
            va_start(arg, INVOKE_TYPE);
            RemoteInvoker     =   va_arg(arg, INVOKER);
            va_end(arg);

            m_mutex = SDL_CreateMutex();
            //m_cond = SDL_CreateCond();
            buf_tid = SDL_CreateThread(WriteThread,0);

            return E_NOERROR;

        case INVOKE_DONELIBRARY:
            SDL_KillThread(buf_tid);
            SDL_DestroyMutex(m_mutex);
            //SDL_DestroyCond(m_cond);

            return E_NOERROR;

        default:
            return E_NOTIMPLEMENTED;
    }
    return E_NOTIMPLEMENTED;
}


