#include "AnsiParser.h"
#include "AnsiList.h"
#include <strings.h>
#include "quickparser.h"

int ReadTo(AnsiParser *P, char *terminator) {
    AnsiString atom;
    do {
        atom=P->NextAtom();
        if (atom==terminator)
            return 1;
        else
        if (atom==(char *)"{")
            ReadTo(P, "}");
        else
        if (atom==(char *)"(")
            ReadTo(P, ")");
        else
        if (atom==(char *)"[")
            ReadTo(P, "]");
    } while (atom!=(char*)"");
    return 0;
}

void AddSortedMember(ClassContainer *CC, ClassMember *CM) {
    int cnt=CC->Members.Count();
    int i;
    for (i=0;i<cnt;i++) {
        ClassMember *CM_O=(ClassMember *)CC->Members.Item(i);
        if (strcasecmp(CM->Name.c_str(),CM_O->Name.c_str())<0)
            break;
    }
    CC->Members.Insert(CM,i,DATA_MEMBER);
}

int FindClass(AnsiList *ClassList, AnsiString item) {
    int count=ClassList->Count();
    for (int i=0;i<count;i++) {
        ClassContainer *CC=(ClassContainer *)ClassList->Item(i);
        if (CC->Name==item)
            return i;
    }
    if (item==(char *)"[]")
        return V_ARRAY;
    return V_NUMBER;
}

int ReadFunctionParams(AnsiParser *P, char *terminator, AnsiString *data, ClassMember *CM, AnsiList *ClassList) {
    AnsiString atom;
    char prec_is_var=0;
    bool first=true;
    char *ptr;
    AnsiString prec;

    do {
        atom=P->NextAtom();
        ptr=atom.c_str();
        if (atom==terminator)
            return 1;
        else
        if (atom==(char*)"var") {
            prec_is_var=1;
        } else {
            if (prec_is_var) {
                prec_is_var=0;
                *data+="var ";
            }

            if ((!first) && (prec)) {
                int id=FindClass(ClassList, prec);

                if (id<0) {
                    id=V_NUMBER;

                    if (prec==(char *)"string") {
                        id=V_NUMBER;
                    } else
                    if (prec==(char *)"array") {
                        id=V_ARRAY;
                    } else
                    if (prec==(char *)"delegate") {
                        id=V_DELEGATE;
                    }
                }
                Variable *var=new Variable;
                var->Name=atom;
                var->Type=id;
                CM->Variables.Add(var,DATA_VARIABLE);

                prec=(char *)"";
            }

            if (*ptr==',') {
                first=true;
                prec=(char *)"";
            } else {
                if (!first) {
                    *data+=" ";
                } else {
                    first=false;
                    prec=atom;
                }
            }

            *data+=atom;
        }
    } while (atom!=(char*)"");
    return 0;
}
/*int ReadFunctionParams(AnsiParser *P, char *terminator, AnsiString *data) {
    AnsiString atom;
    char prec_is_var=0;
    do {
        atom=P->NextAtom();
        if (atom==terminator)
            return 1;
        else
        if (atom==(char*)"var") {
            prec_is_var=1;
        } else {
            if (prec_is_var) {
                prec_is_var=0;
                *data+="var ";
            }
            *data+=atom;
        }
    } while (atom!=(char*)"");
    return 0;
}*/

Variable *FindVar(AnsiList *Variables, AnsiString item) {
    int count=Variables->Count();
    for (int i=0;i<count;i++) {
        Variable *var=(Variable *)Variables->Item(i);
        if (var->Name==item)
            return var;
    }
    return 0;
}

ClassMember *FindMember(AnsiList *Members, AnsiString item) {
    int count=Members->Count();
    for (int i=0;i<count;i++) {
        ClassMember *var=(ClassMember *)Members->Item(i);
        if (var->Name==item)
            return var;
    }
    return 0;
}

int RemoveMember(AnsiList *Members, AnsiString item) {
    int count=Members->Count();
    for (int i=0;i<count;i++) {
        ClassMember *var=(ClassMember *)Members->Item(i);
        if (var->Name==item) {
            Members->Delete(i);
            return 1;
        }
    }
    return 0;
}

int ReadFunction(AnsiParser *P,ClassMember *CM, AnsiList *ClassList,ClassContainer *CC) {
    AnsiString atom;
    char prec_is_var=0;
    char is_assignment=0;
    AnsiString prec;
    AnsiString asg_var;
    int level=0;
    do {
        prec=atom;
        atom=P->NextAtom();
        if ((prec_is_var) && (atom==(char *)"[]")) {
            prec_is_var=2;
        } else
        if (atom==(char *)"{") {
            level++;
        } else
        if (atom==(char *)"}") {
            if (!level)
                return 1;
            else
                level--;
        } else
        if (atom==(char*)"var")
            prec_is_var=1;
        else
        if ((atom==(char*)"=") || (atom==(char*)"=&")) {
            is_assignment=1;
            asg_var=prec;
        } else
        if (is_assignment)  {
            if (atom==(char *)"new") {
                int id=FindClass(ClassList, P->NextAtom());
                Variable *var=FindVar(&CM->Variables, asg_var);
                if (var)
                    var->Type=id;
                else {
                    ClassMember *var_m=FindMember(&CC->Members, asg_var);
                    if (var_m)
                        var_m->Type=id;
                    // cauta in clasa membru ...
                }
            } else {
                Variable *var=FindVar(&CM->Variables, atom);
                int *id=0;
                int def_type=V_NUMBER;

                if (var)
                    id=&var->Type;
                else {

                    ClassMember *var_m=FindMember(&CC->Members, atom);
                    if (var_m) {
                        // verific daca este apel de functie sau delegat
                        if ((var_m->fun_type==1) && (P->NextAtom()!=(char *)"(")) {
                                def_type=V_DELEGATE;
                                id=&def_type;
                        } else
                            id=&var_m->Type;
                    }

                    if (!id) {
                        if ((atom.Length()) && ((atom[0]=='"') || (atom[0]=='\''))) {
                            def_type=V_STRING;
                            id=&def_type;
                        }
                    }
                }

                if (id) {
                    int *id2=0;
                    Variable *var2=FindVar(&CM->Variables, asg_var);
                    if (var2)
                        id2=&var2->Type;//var2->Type=var->Type;
                    else {
                        ClassMember *var_m=FindMember(&CC->Members, asg_var);
                        if (var_m)
                            id2=&var_m->Type;
                    }

                    if (id2)
                        *id2=*id;
                }
            }
            is_assignment=0;
        } else
        if (atom==(char*)"return") {
            AnsiString tmp=P->NextAtom();
            if (tmp==(char *)"new")
                CM->Type=FindClass(ClassList, P->NextAtom());
            else {
                Variable *var=FindVar(&CM->Variables, tmp);
                if (var)
                    CM->Type=var->Type;
                else {
                    if ((tmp.Length()) && ((tmp[0]=='"') || (tmp[0]=='\'')))
                        CM->Type=V_STRING;

                    ClassMember *CM1=FindMember(&CC->Members, tmp);
                    if (CM1) {
                        // este functie ... oare este delegat ?
                        if ((CM1->fun_type==1) && (P->NextAtom()!=(char*)"(")) {
                            CM->Type=V_DELEGATE;
                        } else
                            CM->Type=CM1->Type;
                    } else
                    if (tmp==(char*)"this")
                        CM->Type=FindClass(ClassList, CC->Name);
                }
            }
            //std::cout << "    Function returns : " << "something" << "\n";
        } else
        if (prec_is_var) {
            //std::cout << "    Variable : " << atom << "\n";
            Variable *var=new Variable;
            var->Name=atom;
            if (prec_is_var==1)
                var->Type=V_NUMBER;
            else
                var->Type=V_ARRAY;
            CM->Variables.Add(var,DATA_VARIABLE);

            prec_is_var=0;
        }
    } while (atom!=(char*)"");
    return 0;
}

int DoFunction(AnsiParser *P,ClassContainer *CC, AnsiList *ClassList, char acc) {
    AnsiString atom;
    AnsiString params;
    atom=P->NextAtom();
    //std::cout << "  Function name : " << atom;
    ClassMember *CM=new ClassMember;
    CM->def_start=P->LastLine();
    //CC->Members.Add(CM,DATA_MEMBER);
    CM->Name=atom;
    CM->Type=V_NUMBER;
    CM->fun_type=1;
    CM->access=acc;
    CM->in_include=CC->in_include;
    AddSortedMember(CC,CM);

    atom=P->NextAtom();
    if (atom==(char*)"(") {
        if (!ReadFunctionParams(P,")",&params,CM, ClassList)) {
            CM->def_end=P->LastLine();
            return 0;
        }
    } else {
        CM->def_end=P->LastLine();
        return 0;
    }
    CM->Hint=params;
    //std::cout << ", hint (" << params << ")\n";
    atom=P->NextAtom();

    if (atom==(char*)";") {
        CM->def_end=P->LastLine();
        return 1;
    }
    if (atom!=(char*)"{") {
        CM->def_end=P->LastLine();
        return 0;
    }
    
    int res=ReadFunction(P,CM,ClassList,CC);
    CM->def_end=P->LastLine();
    return res;
}

int DoProperty(AnsiParser *P,ClassContainer *CC, char acc) {
    AnsiString atom;
    atom=P->NextAtom();
    //std::cout << "  Property name : " << atom << "\n";
    ClassMember *CM=new ClassMember();
    CM->def_start=P->LastLine();
    CM->fun_type=2;
    CM->Name=atom;
    CM->Type=V_NUMBER;
    CM->in_include=CC->in_include;
    CM->access=acc;
    //CC->Members.Add(CM,DATA_MEMBER);
    AddSortedMember(CC,CM);

    atom=P->NextAtom();

    if (atom==(char*)";")
        return 1;
    if (atom!=(char*)"{")
        return 0;
    do {
        atom=P->NextAtom();
        if (atom==(char*)"get") {
            atom=P->NextAtom();
            CM->linked_member=atom;
        } else
        if (atom==(char*)"}")
            break;
    } while (atom!=(char *)"");
    //int res=ReadTo(P,"}");
    CM->def_end=P->LastLine();
    //return res;
    return 1;
}

int ExtendClass(ClassContainer *CC_TARGET, AnsiString ClassName, AnsiList *ClassList) {
    int count=ClassList->Count();
    AnsiString result;
    for (int i=0;i<count;i++) {
        ClassContainer *CC=(ClassContainer *)ClassList->Item(i);
        if (CC->Name==ClassName) {
            int count1=CC->Members.Count();
            for (int j=0;j<count1;j++) {
                ClassMember *CM=(ClassMember *)CC->Members.Item(j);
                ClassMember *newCM=new ClassMember;
                newCM->Name=CM->Name;
                newCM->Hint=CM->Hint;
                newCM->def_start=CM->def_start;
                newCM->def_end=CM->def_end;
                newCM->Type=CM->Type;
                newCM->fun_type=CM->fun_type;
                newCM->in_include=CM->in_include;
                newCM->access=CM->access;

                int count3=CM->Variables.Count();
                for (int k=0;k<count3;k++) {
                    Variable *var=(Variable *)CM->Variables.Item(k);
                    Variable *newvar=new Variable;
                    newvar->Name=var->Name;
                    newvar->Name=var->Type;
                    newCM->Variables.Add(newvar, DATA_VARIABLE);
                }

                //CC_TARGET->Members.Add(newCM, DATA_MEMBER);
                AddSortedMember(CC_TARGET,newCM);
            }
            return 1;
        }
    }
    return 0;
}

int GetLinked(AnsiList *Members, char *name, int start_from) {
    int cnt=Members->Count();
    for (int i=start_from;i<cnt;i++) {
        ClassMember *CM=(ClassMember *)Members->Item(i);
        if (CM->Name==name)
            return CM->Type;
    }
    return V_NUMBER;
}

int DoClass(AnsiParser *P, AnsiList *ClassList, char in_include) {
    ClassContainer *CC=new ClassContainer;
    CC->def_start=P->LastLine();
    ClassList->Add(CC,DATA_CLASS);
    AnsiString atom;
    atom=P->NextAtom();
    //std::cout << "Class name : " << atom << "\n";
    CC->Name=atom;
    CC->in_include=in_include;
    char has_prop=0;

    do {
        atom=P->NextAtom();
        if (atom==(char *)"extends") {
            ExtendClass(CC,P->NextAtom(),ClassList);
        } else 
        if (atom==(char*)"{") {
            break;
        } else
            return 0;
    } while (atom!=(char *)"");

    char acces=0;
    char is_static=0;

    do {
        atom=P->NextAtom();
        if (atom==(char *)"public")
            acces=0;
        else
        if (atom==(char *)"private")
            acces=1;
        else
        if (atom==(char *)"protected")
            acces=2;
        else
        if (atom==(char *)"static")
            is_static=1;
        else
        if (atom==(char *)"function") {
            DoFunction(P,CC,ClassList,acces);
            acces=0;
            is_static=0;
        } else
        if (atom==(char *)"property") {
            DoProperty(P,CC,acces);
            has_prop=1;
            acces=0;
            is_static=0;
        } else
        if (atom==(char *)"var") {
            is_static=0;
            atom=P->NextAtom();
            ClassMember *CM=new ClassMember();
            CM->def_start=P->LastLine();
            CM->fun_type=0;
            CM->Name=atom;
            if (atom==(char *)"[]") {
                CM->Name=P->NextAtom();
                CM->Type=V_ARRAY;
            } else
                CM->Type=V_NUMBER;
            CM->def_end=CM->def_start;
            CM->in_include=CC->in_include;
            CM->access=acces;
            //CC->Members.Add(CM,DATA_MEMBER);
            AddSortedMember(CC,CM);
            //std::cout << "  Member variable name : " << atom << "\n";
            acces=0;
        } else
        if (atom==(char *)"event") {
            atom=P->NextAtom();
            //std::cout << "  Event name : " << atom << "\n";

            ClassMember *CM=new ClassMember();
            CM->def_start=P->LastLine();
            CM->fun_type=3;
            CM->Name=atom;
            CM->Type=V_NUMBER;
            CM->in_include=CC->in_include;
            CM->def_end=CM->def_start;
            CM->access=acces;
            //CC->Members.Add(CM,DATA_MEMBER);
            AddSortedMember(CC,CM);

            acces=0;
            is_static=0;
            atom=P->NextAtom();
            if (atom!=(char*)"triggers")
                continue;
            else 
                P->NextAtom();
            CM->def_end=P->LastLine();
        } else
        if (atom==(char *)"override") {
            acces=0;
            is_static=0;
            atom=P->NextAtom();
            RemoveMember(&CC->Members, atom);
            //std::cout << "  Override member : " << atom << "\n";
        }
        if (atom==(char *)"") {
            CC->def_end=P->LastLine();
            return 0;
        }

    } while (atom!=(char*)"}");
    CC->def_end=P->LastLine();
    if (has_prop) {
        int cnt=CC->Members.Count();
        for (int i=0;i<cnt;i++) {
            ClassMember *CM1=(ClassMember *)CC->Members.Item(i);
            if (CM1->linked_member.Length())
                CM1->Type=GetLinked(&CC->Members,CM1->linked_member.c_str(),0);
        }
    }
    return 1;
}

AnsiList ClassList;

int Parse(void *owner, AnsiString *text, INCLUDE_REQUEST cback, char in_include, char clean) {
    AnsiList *TargetList=&ClassList;
    int index=0;

    int cnt=TargetList->Count();
    for (int i=0;i<cnt;i++) {
        ClassContainer *CC=(ClassContainer *)TargetList->Item(index);
        if (!CC->in_include)
            TargetList->Delete(index);
        else
            index++;
    }

    if (clean) {
        TargetList->Clear();
    }

    AnsiList ConstantsList;
    //AnsiList ClassList;
    AnsiParser P(text,&ConstantsList);
    AnsiString atom;
    do {
        atom=P.NextAtom();

        if (atom==(char *)"class")
            DoClass(&P,TargetList, in_include);
        else
        if (atom==(char *)"pragma")
            P.GetConstant();
        else
        if (atom==(char *)"define")
            P.GetConstant();
        else
        if (atom==(char *)"import")
            P.GetConstant();
        else
        if (atom==(char *)"include") {
            atom=P.GetConstant();
            if (cback) {
                //int start=IncludeClassList.Count();
                cback(owner, Strip(atom).c_str());
                //AddToMain(&IncludeClassList, &ClassList, start);
            }
        }

    } while (atom!=(char*)"");
    //PrintData(&ClassList);
    return 0;
}

int PreParse(void *owner, AnsiString *text, INCLUDE_REQUEST cback) {
    //AnsiString text;
    //text.LoadFile("test.con");
    AnsiList ConstantsList;
    //AnsiList ClassList;
    AnsiParser P(text,&ConstantsList);
    AnsiString atom;
    do {
        atom=P.NextAtom();

        if (atom==(char *)"include") {
            atom=P.GetConstant();
            if (cback)
                cback(owner, Strip(atom).c_str());
        }
    } while (atom!=(char*)"");
    //PrintData(&ClassList);
    return 0;
}

int Sort(AnsiList *lst, char *pelem) {
    int count=lst->Count();
    for (int i=0;i<count;i++) {
        if (strcasecmp(((AnsiString *)lst->Item(i))->c_str(),pelem)>0)
            return i;
    }
    return count;
}

AnsiString GetSciList(AnsiString ClassName) {
    int count=ClassList.Count();
    AnsiString result;
    for (int i=0;i<count;i++) {
        ClassContainer *CC=(ClassContainer *)ClassList.Item(i);
        if (CC->Name==ClassName) {
            int count2=CC->Members.Count();
            AnsiList temp_list_public;
            AnsiList temp_list_not_public;
            
            AnsiString element;
            for (int j=0;j<count2;j++) {
                ClassMember *CM=(ClassMember *)CC->Members.Item(j);
                // the list isn't an "autoclean" list
                AnsiString *element=new AnsiString(CM->Name.c_str());
                *element+="?";
                *element+=AnsiString((long)(CM->access ? (CM->fun_type+11) : (CM->fun_type+1) ));

                if (CM->access==0)
                    temp_list_public.Insert(element, Sort(&temp_list_public,CM->Name.c_str()), DATA_STRING);
                else
                    temp_list_not_public.Insert(element, Sort(&temp_list_not_public,CM->Name.c_str()), DATA_STRING);
                /*if (j)
                    result+=" ";
                result+=CM->Name;*/
            }
            int cnt=temp_list_public.Count();
            for (int k=0;k<count2;k++) {
                if (k)
                    result+=(char *)" ";
                if (k<cnt)
                    result+=*(AnsiString *)temp_list_public.Item(k);
                else {
                    result+=*(AnsiString *)temp_list_not_public.Item(k-cnt);
                }
            }
            break;
        }
    }
    return result;
}

AnsiString GetHint(AnsiString ClassName, AnsiString Member) {
    int count=ClassList.Count();
    AnsiString result;
    for (int i=0;i<count;i++) {
        ClassContainer *CC=(ClassContainer *)ClassList.Item(i);
        if (CC->Name==ClassName) {
            int count2=CC->Members.Count();
            for (int j=0;j<count2;j++) {
                ClassMember *CM=(ClassMember *)CC->Members.Item(j);
                if (CM->Name==Member) {
                    if (CM->fun_type!=1) {
                        result="Warning: This is not a member function";
                        return result;
                    }
                    switch (CM->Type) {
                        case V_DELEGATE :
                            result="delegate ";
                            break;
                        case V_ARRAY    :
                            result="[] ";
                            break;
                        case V_STRING   :
                            result="string ";
                            break;
                        case V_NUMBER   :
                            result="number ";
                            break;
                        default:
                            {
                                ClassContainer *CC1=(ClassContainer *)ClassList.Item(CM->Type);
                                if (CC1)
                                    result=CC1->Name;
                                result+=(char *)" ";
                            }
                    }
                    AnsiString hint=(char *)"(";
                    hint+=CM->Hint;
                    hint+=(char *)")";
                    result+=ClassName;
                    result+=(char *)".";
                    result+=Member;
                    result+=hint;
                    result=(hint+"\n\nDeclared as:\n")+result;
                }
            }
            break;
        }
    }
    return result;
}

//#include <iostream>
AnsiString GetElementType(AnsiString element, int line) {
    int count=ClassList.Count();
    AnsiString result;
    for (int i=0;i<count;i++) {
        ClassContainer *CC=(ClassContainer *)ClassList.Item(i);
        if ((!CC->in_include) && (line>=CC->def_start) && (line<=CC->def_end)) {
            if (element==(char *)"this") {
                result=CC->Name;
                break;
            }
            int count2=CC->Members.Count();
            for (int j=0;j<count2;j++) {
                ClassMember *CM=(ClassMember *)CC->Members.Item(j);
                if ((!CM->in_include) && (line>=CM->def_start) && (line<=CM->def_end)) {
                    int count3=CM->Variables.Count();
                    for (int k=0;k<count3;k++) {
                        Variable *var=(Variable*)CM->Variables.Item(k);
                        if (var->Name==element) {
                            if (var->Type>=0) {
                                ClassContainer *CC1=(ClassContainer *)ClassList.Item(var->Type);
                                if (CC1) {
                                    result=CC1->Name;
                                }
                            }
                            break;
                        }
                    }
                    break;
                }
            }
            break;
        }
    }
    return result;
}

AnsiString GetMemberType(AnsiString ClassName, AnsiString element) {
    int count=ClassList.Count();
    AnsiString result;
    for (int i=0;i<count;i++) {
        ClassContainer *CC=(ClassContainer *)ClassList.Item(i);
        if (CC->Name==ClassName) {
            int count2=CC->Members.Count();
            for (int j=0;j<count2;j++) {
                ClassMember *CM=(ClassMember *)CC->Members.Item(j);
                if (CM->Name==element) {
                    if (CM->Type>=0) {
                        ClassContainer *CC1=(ClassContainer *)ClassList.Item(CM->Type);
                        if (CC1) {
                            result=CC1->Name;
                        }
                    }
                    break;
                }
            }
            break;
        }
    }
    return result;
}

AnsiString DoType(int Type) {
        switch (Type) {
            case V_DELEGATE :
                return "delegate";
                break;
            case V_ARRAY    :
                return "[]";
                break;
            case V_STRING   :
                return "string";
                break;
            case V_NUMBER   :
                return "number";
                break;
            default:
                {
                    ClassContainer *CC1=(ClassContainer *)ClassList.Item(Type);
                    if (CC1)
                        return CC1->Name;
                }
        }
        return "number";
}

AnsiString LastProjectList;

AnsiString GetProjectList() {
    int count=ClassList.Count();
    AnsiString result;
    char first_class=1;
    for (int i=0;i<count;i++) {
        ClassContainer *CC=(ClassContainer *)ClassList.Item(i);
        if (!CC->in_include) {
            int count2=CC->Members.Count();
            char first;
            if (count2) {

                if (!first_class)
                    result+=(char *)"|";
                else
                    first_class=0;

                result+=CC->Name;

                result+=(char *)":";

                first=1;
                for (int j=0;j<count2;j++) {
                    ClassMember *CM=(ClassMember *)CC->Members.Item(j);
                    if (!CM->in_include) {
                        if (!first) {
                            result+=(char*)"%";
                        } else
                            first=0;

                        result+=AnsiString((long)CM->access);
                        result+=(char*)";";
                        result+=AnsiString((long)CM->fun_type);
                        result+=(char*)";";
                        result+=DoType(CM->Type);
                        result+=(char*)";";
                        result+=CM->Name;
                        result+=(char*)";";
                        result+=CM->Hint;
                        result+=(char*)";";
                        result+=AnsiString((long)CM->def_start);
                    }
                }
            }
        }
    }
    return result;
}

char ProjectListChanged() {
    AnsiString temp=GetProjectList();
    int res=LastProjectList!=temp;
    if (res)
        LastProjectList=temp;
    return res;
}

AnsiString GetCheckedProjectList() {
    return LastProjectList;
}
