//------------ standard header -----------------------------------//
#include "stdlibrary.h"
//------------ end of standard header ----------------------------//
#include <stdarg.h>
#ifdef _WIN32
	#include <windows.h>
    #define THREAD_TYPE DWORD WINAPI
#else
    #include <pthread.h>
    #define THREAD_TYPE LPVOID
    #define LPVOID void*
#endif

#include <gtk/gtk.h>
#include "AnsiString.h"
#include <unistd.h>
#include "semhh.h"

#define CV_NO_BACKWARD_COMPATIBILITY
#include <cv.h>
#include <highgui.h>

#define CUSTOM_CONTROL_ID       1015

#define P_USEWIDGET             1000
#define P_UPDATEWIDGET          1001
#define P_UPDATEFRAME           1002
#define P_CAPTUREINIT           1003
#define P_RELEASECAPTURE        1004
#define P_CAPTUREFORMAT         1005

#define P_VIDEOFORMAT           1006
#define P_VIDEOWRITERCREATE     1007
#define P_VIDEOWRITERDONE       1008
#define P_VIDEOUPDATE           1009
#define P_VIDEOFPS              1010
#define P_VIDEOINIT             1011
#define P_CLEARSTREAMDATA       1012
#define P_VIDEOUPDATEIMAGE      1014
#define P_STREAMDATA            1015
#define P_MOTIONDETECTOR        1016
#define P_MOTIONTRESHOLD        1017
#define P_USEWIDGET2            1018
#define P_UPDATEMOTION          1019

#define P_CLEAROBJECTS          1020
#define P_SCALEOBJECTS          1021
#define P_UPDATEOBJECT          1022
#define P_USEWIDGET3            1023

#define P_TRACKVMIN             1024
#define P_TRACKVMAX             1025
#define P_TRACKSMIN             1026
#define P_USEWIDGET4            1027
#define P_UPDATETRACK           1028
#define P_TRACKOBJECT           1029
#define P_LASTHISTOGRAM         1030

#define P_FRAMEWIDTH            1031
#define P_FRAMEHEIGHT           1032

#define EVENT_ON_MOTION         350
#define EVENT_ON_OBJECTDETECTED 351
#define EVENT_ON_TRACK          352

using namespace cv;
using namespace std;

INVOKER RemoteInvoker=0;
AnsiString buffered_text;

#define MAX_PAIRS   0xFF
struct CascadePair {
    CascadeClassifier cascade;
    CascadeClassifier nested;
};

class TrackContext {
    public:
        IplImage *image;
        IplImage *hsv;
        IplImage *hue;
        IplImage *mask;
        IplImage *backproject;
        IplImage *histimg;
        int backproject_mode;
        int select_object;
        int track_object;
        int show_hist;
        CvPoint origin;
        CvRect selection;
        CvRect track_window;
        CvBox2D track_box;
        CvConnectedComp track_comp;
        int hdims;
        float hranges_arr[2];
        float* hranges;
        CvHistogram *hist;
        int vmin;
        int vmax;
        int smin;
        bool track_event;

        TrackContext() {
            image = 0;
            hsv = 0;
            hue = 0;
            mask = 0;
            backproject = 0;
            histimg = 0;
            backproject_mode = 0;
            select_object = 0;
            track_object = 0;
            show_hist = 1;
            hdims = 16;
            hranges_arr[0]=0;
            hranges_arr[1]=180;
            hranges = hranges_arr;
            hist=0;
            vmin = 10;
            vmax = 256;
            smin = 30;
            track_event = false;
        }
};

THREAD_TYPE ProcThread(LPVOID SENDER);
struct CaptureContext {
    CvCapture* ctx;
    GtkWidget* renderto;
    GtkWidget* secondary_renderto;
    GtkWidget* object_renderto;
    GtkWidget* track_renderto;
    AnsiString format;
    AnsiString video_format;
    CvVideoWriter *video_writer;
    AnsiString streamname;
    AnsiString stream_data;
    int fps;
    bool motion_detector;
    bool motion_event;
    int motion_treshold;

    // ring image buffer
    IplImage **buf;
    IplImage *motion;
    int last;

    CvMemStorage* storage; // temporary storage
    // temporary images
    IplImage *mhi; // MHI
    IplImage *orient; // orientation
    IplImage *mask; // valid orientation mask
    IplImage *segmask; // motion segmentation map

    // face detection
    bool detect_event;
    double scale;
    CascadePair cpair[MAX_PAIRS];
    int cpair_len;


    AnsiString img_name;
    char *img_name_ptr;
    TrackContext tc;
};

void *pending_image;
char *p_target_name;
HHSEM lock;
bool is_done;
bool use_imwrite;

CaptureContext* MasterCtx=0;

// This is done in a thread, because on the main thread it uses to crash on windows 7
// when i move the main window or resize it (when glib::on_idle is affected).
// When running cvSaveImage in a thread, everything seems right.
//
// This is the gdb output:
// gdb: unknown target exception 0xc0000008 at 0x777a64cf
// Program received signal ?, Unknown signal.
// 0x777a64cf in ntdll!LdrFlushAlternateResourceModules ()

THREAD_TYPE ProcThread(LPVOID SENDER) {
    while (true) {
        semp(lock);
        if ((pending_image) && (pending_image)) {
            if (use_imwrite)
                imwrite(p_target_name, *(Mat *)pending_image);
            else
                cvSaveImage(p_target_name, (IplImage *)pending_image);
            pending_image=0;
            p_target_name=0;
            is_done=true;
            semv(lock);
        } else {
            semv(lock);
#ifdef _WIN32
            Sleep(1);
#else
            usleep(1000);
#endif
        }

    }
    return 0;
}
/*
void MySaveImage(char *fname, void *frame, bool _use_imwrite=false) {
    semp(lock);
    is_done=false;
    p_target_name=fname;
    pending_image=frame;
    use_imwrite=_use_imwrite;
    semv(lock);
    int stay_in=true;
    while (stay_in) {
        semp(lock);
        if (is_done)
            stay_in=false;
        semv(lock);
        if (stay_in)
#ifdef _WIN32
            Sleep(1);
#else
            usleep(1000);
#endif

    }
}

void InitProcThread() {
    seminit(lock, 1);
#ifdef _WIN32
    DWORD TID;
    CreateThread(0,0,ProcThread,0,0,&TID);
#else
    pthread_t threadID;
    pthread_create(&threadID, NULL, ProcThread,0);
#endif
}*/

CvScalar hsv2rgb( float hue )
{
    int rgb[3], p, sector;
    static const int sector_data[][3]=
        {{0,2,1}, {1,2,0}, {1,0,2}, {2,0,1}, {2,1,0}, {0,1,2}};
    hue *= 0.033333333333333333333333333333333f;
    sector = cvFloor(hue);
    p = cvRound(255*(hue - sector));
    p ^= sector & 1 ? 255 : 0;

    rgb[sector_data[sector][0]] = 255;
    rgb[sector_data[sector][1]] = 0;
    rgb[sector_data[sector][2]] = p;

    return cvScalar(rgb[2], rgb[1], rgb[0],0);
}

void track_object(CaptureContext *ctx, IplImage* frame) {
    if( !ctx->tc.image )
    {
        /* allocate all the buffers */
        ctx->tc.image = cvCreateImage( cvGetSize(frame), 8, 3 );
        ctx->tc.image->origin = frame->origin;
        ctx->tc.hsv = cvCreateImage( cvGetSize(frame), 8, 3 );
        ctx->tc.hue = cvCreateImage( cvGetSize(frame), 8, 1 );
        ctx->tc.mask = cvCreateImage( cvGetSize(frame), 8, 1 );
        ctx->tc.backproject = cvCreateImage( cvGetSize(frame), 8, 1 );
        ctx->tc.hist = cvCreateHist( 1, &ctx->tc.hdims, CV_HIST_ARRAY, &ctx->tc.hranges, 1 );
        ctx->tc.histimg = cvCreateImage( cvSize(320,200), 8, 3 );
        cvZero( ctx->tc.histimg );
    }
    cvCopy( frame, ctx->tc.image, 0 );
    cvCvtColor( ctx->tc.image, ctx->tc.hsv, CV_BGR2HSV );

    if( ctx->tc.track_object )
    {
        int _vmin = ctx->tc.vmin, _vmax = ctx->tc.vmax;

        cvInRangeS( ctx->tc.hsv, cvScalar(0,ctx->tc.smin,MIN(_vmin,_vmax),0),
                    cvScalar(180,256,MAX(_vmin,_vmax),0), ctx->tc.mask );
        cvSplit( ctx->tc.hsv, ctx->tc.hue, 0, 0, 0 );

        if( ctx->tc.track_object < 0 )
        {
            //fprintf(stderr, "HERE ! X:%i Y:%i W:%i H:%i\n",ctx->tc.selection.x, ctx->tc.selection.y, ctx->tc.selection.width, ctx->tc.selection.height);
            float max_val = 0.f;
            cvSetImageROI( ctx->tc.hue, ctx->tc.selection );
            cvSetImageROI( ctx->tc.mask, ctx->tc.selection );
            cvCalcHist( &ctx->tc.hue, ctx->tc.hist, 0, ctx->tc.mask );
            cvGetMinMaxHistValue( ctx->tc.hist, 0, &max_val, 0, 0 );
            cvConvertScale( ctx->tc.hist->bins, ctx->tc.hist->bins, max_val ? 255. / max_val : 0., 0 );
            cvResetImageROI( ctx->tc.hue );
            cvResetImageROI( ctx->tc.mask );
            ctx->tc.track_window = ctx->tc.selection;
            ctx->tc.track_object = 1;

            cvZero( ctx->tc.histimg );
            int bin_w = ctx->tc.histimg->width / ctx->tc.hdims;
            for( int i = 0; i < ctx->tc.hdims; i++ )
            {
                int val = cvRound( cvGetReal1D(ctx->tc.hist->bins,i)*ctx->tc.histimg->height/255 );
                CvScalar color = hsv2rgb(i*180.f/ctx->tc.hdims);
                cvRectangle( ctx->tc.histimg, cvPoint(i*bin_w,ctx->tc.histimg->height),
                             cvPoint((i+1)*bin_w,ctx->tc.histimg->height - val),
                             color, -1, 8, 0 );
            }
        }

        cvCalcBackProject( &ctx->tc.hue, ctx->tc.backproject, ctx->tc.hist );
        cvAnd( ctx->tc.backproject, ctx->tc.mask, ctx->tc.backproject, 0 );
        cvCamShift( ctx->tc.backproject, ctx->tc.track_window,
                    cvTermCriteria( CV_TERMCRIT_EPS | CV_TERMCRIT_ITER, 10, 1 ),
                    &ctx->tc.track_comp, &ctx->tc.track_box );
        ctx->tc.track_window = ctx->tc.track_comp.rect;

        if( ctx->tc.backproject_mode )
            cvCvtColor( ctx->tc.backproject, ctx->tc.image, CV_GRAY2BGR );
        if( !ctx->tc.image->origin )
            ctx->tc.track_box.angle = -ctx->tc.track_box.angle;
        cvEllipseBox( ctx->tc.image, ctx->tc.track_box, CV_RGB(255,0,0), 3, CV_AA, 0 );

        if (ctx->tc.track_event) {
            AnsiString track_list("<list>");
            track_list+="<o rad=\"";
            track_list+=AnsiString((long)ctx->tc.track_box.size.width/2);
            track_list+="\" rad2=\"";
            track_list+=AnsiString((long)ctx->tc.track_box.size.height/2);
            track_list+="\" x=\"";
            track_list+=AnsiString((long)ctx->tc.track_box.center.x);
            track_list+="\" y=\"";
            track_list+=AnsiString((long)ctx->tc.track_box.center.y);
            track_list+="\" angle=\"";
            track_list+=AnsiString((double)ctx->tc.track_box.angle);
            track_list+="\" />\n";
            track_list+="</list>";
            RemoteInvoker(ctx, INVOKE_FIREEVENT, (int)EVENT_ON_TRACK, track_list.c_str());
        }

    }

    /*if( ctx->tc.select_object && ctx->tc.selection.width > 0 && ctx->tc.selection.height > 0 )
    {
        cvSetImageROI( ctx->tc.image, ctx->tc.selection );
        cvXorS( ctx->tc.image, cvScalarAll(255), ctx->tc.image, 0 );
        cvResetImageROI( ctx->tc.image );
    } */
}

void  update_mhi(CaptureContext *ctx, IplImage* img)
{
    int diff_threshold = ctx->motion_treshold;
    // various tracking parameters (in seconds)
    const double MHI_DURATION = 1;
    const double MAX_TIME_DELTA = 0.5;
    const double MIN_TIME_DELTA = 0.05;
    // number of cyclic frame buffer used for motion detection
    // (should, probably, depend on FPS)
    const int N = 4;

    double timestamp = (double)clock()/CLOCKS_PER_SEC; // get current time in seconds
    CvSize size = cvSize(img->width,img->height); // get current frame size
    int i, idx1 = ctx->last, idx2;
    IplImage* silh;
    CvSeq* seq;
    CvRect comp_rect;
    double count;
    double angle;
    CvPoint center;
    double magnitude;
    CvScalar color;
    AnsiString motion_list("<list>\n");
    int has_motion=false;

    if (!ctx->motion) {
        ctx->motion = cvCreateImage( cvSize(img->width,img->height), 8, 3 );
        cvZero( ctx->motion ); 
        ctx->motion->origin = img->origin;
    }
    // allocate images at the beginning or
    // reallocate them if the frame size is changed
    if( !ctx->mhi || ctx->mhi->width != size.width || ctx->mhi->height != size.height ) {
        if( ctx->buf == 0 ) {
            ctx->buf = (IplImage**)malloc(N*sizeof(ctx->buf[0]));
            memset( ctx->buf, 0, N*sizeof(ctx->buf[0]));
        }

        for( i = 0; i < N; i++ ) {
            cvReleaseImage( &ctx->buf[i] );
            ctx->buf[i] = cvCreateImage( size, IPL_DEPTH_8U, 1 );
            cvZero( ctx->buf[i] );
        }
        cvReleaseImage( &ctx->mhi );
        cvReleaseImage( &ctx->orient );
        cvReleaseImage( &ctx->segmask );
        cvReleaseImage( &ctx->mask );

        ctx->mhi = cvCreateImage( size, IPL_DEPTH_32F, 1 );
        cvZero( ctx->mhi ); // clear MHI at the beginning
        ctx->orient = cvCreateImage( size, IPL_DEPTH_32F, 1 );
        ctx->segmask = cvCreateImage( size, IPL_DEPTH_32F, 1 );
        ctx->mask = cvCreateImage( size, IPL_DEPTH_8U, 1 );
    }

    cvCvtColor( img, ctx->buf[ctx->last], CV_BGR2GRAY ); // convert frame to grayscale

    idx2 = (ctx->last + 1) % N; // index of (last - (N-1))th frame
    ctx->last = idx2;

    silh = ctx->buf[idx2];
    cvAbsDiff( ctx->buf[idx1], ctx->buf[idx2], silh ); // get difference between frames

    cvThreshold( silh, silh, diff_threshold, 1, CV_THRESH_BINARY ); // and threshold it
    cvUpdateMotionHistory( silh, ctx->mhi, timestamp, MHI_DURATION ); // update MHI

    // convert MHI to blue 8u image
    cvCvtScale( ctx->mhi, ctx->mask, 255./MHI_DURATION,
                (MHI_DURATION - timestamp)*255./MHI_DURATION );
    cvZero( ctx->motion );
    cvMerge( ctx->mask, 0, 0, 0, ctx->motion );

    // calculate motion gradient orientation and valid orientation mask
    cvCalcMotionGradient( ctx->mhi, ctx->mask, ctx->orient, MAX_TIME_DELTA, MIN_TIME_DELTA, 3 );

    if( !ctx->storage )
        ctx->storage = cvCreateMemStorage(0);
    else
        cvClearMemStorage(ctx->storage);

    // segment motion: get sequence of motion components
    // segmask is marked motion components map. It is not used further
    seq = cvSegmentMotion( ctx->mhi, ctx->segmask, ctx->storage, timestamp, MAX_TIME_DELTA );

    // iterate through the motion components,
    // One more iteration (i == -1) corresponds to the whole image (global motion)
    for( i = -1; i < seq->total; i++ ) {

        if( i < 0 ) { // case of the whole image
            comp_rect = cvRect( 0, 0, size.width, size.height );
            color = CV_RGB(255,255,255);
            magnitude = 100;
        }
        else { // i-th motion component
            comp_rect = ((CvConnectedComp*)cvGetSeqElem( seq, i ))->rect;
            if( comp_rect.width + comp_rect.height < 100 ) // reject very small components
                continue;
            color = CV_RGB(255,0,0);
            magnitude = 30;
        }

        // select component ROI
        cvSetImageROI( silh, comp_rect );
        cvSetImageROI( ctx->mhi, comp_rect );
        cvSetImageROI( ctx->orient, comp_rect );
        cvSetImageROI( ctx->mask, comp_rect );

        // calculate orientation
        angle = cvCalcGlobalOrientation( ctx->orient, ctx->mask, ctx->mhi, timestamp, MHI_DURATION);
        angle = 360.0 - angle;  // adjust for images with top-left origin

        count = cvNorm( silh, 0, CV_L1, 0 ); // calculate number of points within silhouette ROI

        cvResetImageROI( ctx->mhi );
        cvResetImageROI( ctx->orient );
        cvResetImageROI( ctx->mask );
        cvResetImageROI( silh );

        // check for the case of little motion
        if( count < comp_rect.width*comp_rect.height * 0.05 )
            continue;

        // draw a clock with arrow indicating the direction
        center = cvPoint( (comp_rect.x + comp_rect.width/2),
                          (comp_rect.y + comp_rect.height/2) );

        if (ctx->motion_event) {     
            motion_list+="<o mag=\"";
            motion_list+=AnsiString(magnitude);
            motion_list+="\" x=\"";
            motion_list+=AnsiString((long)center.x);
            motion_list+="\" y=\"";
            motion_list+=AnsiString((long)center.y);
            motion_list+="\" angle=\"";
            motion_list+=AnsiString((double)angle);
            motion_list+="\" />\n";
            has_motion=true;
        }
        cvCircle( ctx->motion, center, cvRound(magnitude*1.2), color, 3, CV_AA, 0 );
        cvLine( ctx->motion, center, cvPoint( cvRound( center.x + magnitude*cos(angle*CV_PI/180)),
                cvRound( center.y - magnitude*sin(angle*CV_PI/180))), color, 3, CV_AA, 0 );
    }
    if ((ctx->motion_event) && (has_motion)) {
        motion_list+="</list>";
        RemoteInvoker(ctx, INVOKE_FIREEVENT, (int)EVENT_ON_MOTION, motion_list.c_str());
    }
}

void detectAndDraw( CaptureContext *ctx, Mat& img) {
    int i = 0;
    //double t = 0;
    vector<Rect> faces;
    CascadeClassifier *cascade;
    CascadeClassifier *nestedCascade;
    bool has_object=false;
    AnsiString object_list("<list>\n");
    const static Scalar colors[] =  { CV_RGB(0,0,255),
        CV_RGB(0,128,255),
        CV_RGB(0,255,255),
        CV_RGB(0,255,0),
        CV_RGB(255,128,0),
        CV_RGB(255,255,0),
        CV_RGB(255,0,0),
        CV_RGB(255,0,255)} ;
    if (!ctx->scale)
        ctx->scale=1;
    Mat gray, smallImg( cvRound (img.rows/ctx->scale), cvRound(img.cols/ctx->scale), CV_8UC1 );

    cvtColor( img, gray, CV_BGR2GRAY );
    resize( gray, smallImg, smallImg.size(), 0, 0, INTER_LINEAR );
    equalizeHist( smallImg, smallImg );

    //t = (double)cvGetTickCount();
    for (int index=0;index<ctx->cpair_len;index++) {
        cascade = &ctx->cpair[i].cascade;
        nestedCascade = &ctx->cpair[i].nested;
        if (cascade) {
            cascade->detectMultiScale( smallImg, faces,
                1.1, 2, 0
                //|CV_HAAR_FIND_BIGGEST_OBJECT
                //|CV_HAAR_DO_ROUGH_SEARCH
                |CV_HAAR_SCALE_IMAGE
                ,
                Size(30, 30) );
        }
        //t = (double)cvGetTickCount() - t;
        //printf( "detection time = %g ms\n", t/((double)cvGetTickFrequency()*1000.) );
        for( vector<Rect>::const_iterator r = faces.begin(); r != faces.end(); r++, i++ )
        {
            Mat smallImgROI;
            vector<Rect> nestedObjects;
            Point center;
            Scalar color = colors[i%8];
            int radius;
            center.x = cvRound((r->x + r->width*0.5)*ctx->scale);
            center.y = cvRound((r->y + r->height*0.5)*ctx->scale);
            radius = cvRound((r->width + r->height)*0.25*ctx->scale);
            circle( img, center, radius, color, 3, 8, 0 );

            if (ctx->detect_event) {
                has_object=true;

                object_list+="<o rad=\"";
                object_list+=AnsiString(radius);
                object_list+="\" x=\"";
                object_list+=AnsiString((long)center.x);
                object_list+="\" y=\"";
                object_list+=AnsiString((long)center.y);
                if ( ( !nestedCascade ) || ( nestedCascade->empty() ) ) {
                    object_list+="\" />\n";
                    continue;
                } else
                    object_list+="\">\n";
            }

            if ( ( !nestedCascade ) || ( nestedCascade->empty() ) )
                continue;
            smallImgROI = smallImg(*r);
            nestedCascade->detectMultiScale( smallImgROI, nestedObjects,
                1.1, 2, 0
                //|CV_HAAR_FIND_BIGGEST_OBJECT
                //|CV_HAAR_DO_ROUGH_SEARCH
                //|CV_HAAR_DO_CANNY_PRUNING
                |CV_HAAR_SCALE_IMAGE
                ,
                Size(30, 30) );
            for( vector<Rect>::const_iterator nr = nestedObjects.begin(); nr != nestedObjects.end(); nr++ )
            {
                center.x = cvRound((r->x + nr->x + nr->width*0.5)*ctx->scale);
                center.y = cvRound((r->y + nr->y + nr->height*0.5)*ctx->scale);
                radius = cvRound((nr->width + nr->height)*0.25*ctx->scale);
                circle( img, center, radius, color, 3, 8, 0 );

                if (ctx->detect_event) {
                    object_list+="<o rad=\"";
                    object_list+=AnsiString(radius);
                    object_list+="\" x=\"";
                    object_list+=AnsiString((long)center.x);
                    object_list+="\" y=\"";
                    object_list+=AnsiString((long)center.y);
                    object_list+="\" />\n";
                }
            }
            object_list+="</o>\n";
        }  
    }
    if ((ctx->detect_event) && (has_object)) {
        object_list+="</list>";
        RemoteInvoker(ctx, INVOKE_FIREEVENT, (int)EVENT_ON_OBJECTDETECTED, object_list.c_str());
    }
}

PREFIX int Invoke(void *context,int INVOKE_TYPE,...) {
    va_list arg;
    switch (INVOKE_TYPE) {
        case INVOKE_QUERYTEST:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id=va_arg(arg, int);
                va_end(arg);

                if (cls_id==CUSTOM_CONTROL_ID)
                    return E_ACCEPTED;
            }
            break;
        case INVOKE_SETEVENT:
            {
                if (!RemoteInvoker)
                    break;
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int event_id    =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                CaptureContext *ctx=(CaptureContext *)context;
                switch (event_id) {
                    case EVENT_ON_MOTION:
                        ctx->motion_event=true;
                        return E_ACCEPTED;
                        break;
                    case EVENT_ON_OBJECTDETECTED:
                        ctx->detect_event=true;
                        return E_ACCEPTED;
                        break;
                    case EVENT_ON_TRACK:
                        ctx->tc.track_event=true;
                        return E_ACCEPTED;
                        break;
                }
            }
            break;
        case INVOKE_SETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char *str_buffer=   va_arg(arg, char *);
                int buf_len     =   va_arg(arg, int);
                va_end(arg);
                AnsiString param(str_buffer);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                CaptureContext *ctx=(CaptureContext *)context;
                switch (prop_id) {
                    case P_USEWIDGET:
                        if (ctx->ctx) {
                            int ctrl_id=param.ToInt();
                            int cls_id=0;
                            if (ctrl_id>0) {
                                RemoteInvoker(context,INVOKE_GETCONTROL, ctrl_id, (void **)&ctx->renderto, &cls_id);
                                if (cls_id!=18)
                                    ctx->renderto=0;
                            } else {
                                ctx->renderto=0;
                            }
                        }
                        break;
                    case P_USEWIDGET2:
                        if (ctx->ctx) {
                            int ctrl_id=param.ToInt();
                            int cls_id=0;
                            if (ctrl_id>0) {
                                RemoteInvoker(context,INVOKE_GETCONTROL, ctrl_id, (void **)&ctx->secondary_renderto, &cls_id);
                                if (cls_id!=18)
                                    ctx->renderto=0;
                            } else {
                                ctx->renderto=0;
                            }
                        }
                        break;
                    case P_USEWIDGET3:
                        if (ctx->ctx) {
                            int ctrl_id=param.ToInt();
                            int cls_id=0;
                            if (ctrl_id>0) {
                                RemoteInvoker(context,INVOKE_GETCONTROL, ctrl_id, (void **)&ctx->object_renderto, &cls_id);
                                if (cls_id!=18)
                                    ctx->object_renderto=0;
                            } else {
                                ctx->object_renderto=0;
                            }
                        }
                        break;
                    case P_USEWIDGET4:
                        if (ctx->ctx) {
                            int ctrl_id=param.ToInt();
                            int cls_id=0;
                            if (ctrl_id>0) {
                                RemoteInvoker(context,INVOKE_GETCONTROL, ctrl_id, (void **)&ctx->track_renderto, &cls_id);
                                if (cls_id!=18)
                                    ctx->track_renderto=0;
                            } else {
                                ctx->track_renderto=0;
                            }
                        }
                        break;
                    case P_UPDATEFRAME:
                        break;
                    case P_CAPTUREFORMAT:
                        {
                            ctx->format=param;
                            ctx->img_name=(char *)"cv_image";
                            ctx->img_name+=AnsiString((long)ctx);
                            ctx->img_name+=AnsiString("_");
                            ctx->img_name+=AnsiString((long)ctx->renderto);
                            ctx->img_name+=(char *)".";
                            ctx->img_name+=ctx->format;
                            ctx->img_name_ptr=ctx->img_name.c_str();
                        }
                        break;
                    case P_UPDATEWIDGET:
                        if (ctx->ctx) {
                            if ((ctx->renderto) || (ctx->motion_detector)) {
                                IplImage* frame = 0;
                                frame = cvQueryFrame( ctx->ctx );
                                if (frame) {
                                    if (!ctx->img_name_ptr) {
                                        ctx->img_name=(char *)"cv_image";
                                        ctx->img_name+=AnsiString((long)ctx);
                                        ctx->img_name+=AnsiString("_");
                                        ctx->img_name+=AnsiString((long)ctx->renderto);
                                        ctx->img_name+=(char *)".";
                                        ctx->img_name+=ctx->format;
                                        ctx->img_name_ptr=ctx->img_name.c_str();
                                    }
                                    if (ctx->renderto) {
                                        cvSaveImage(ctx->img_name_ptr, frame);
                                        gtk_image_set_from_file(GTK_IMAGE(ctx->renderto), ctx->img_name_ptr);
                                        unlink(ctx->img_name_ptr);
                                    }
                                    if (ctx->tc.track_object) {
                                        track_object(ctx, frame);
                                        if (ctx->tc.image) {
                                            if (ctx->track_renderto) {
                                                cvSaveImage(ctx->img_name_ptr, ctx->tc.image);
                                                gtk_image_set_from_file(GTK_IMAGE(ctx->track_renderto), ctx->img_name_ptr);
                                                unlink(ctx->img_name_ptr);
                                            }
                                        }
                                    }
                                    if (ctx->cpair_len) {
                                        Mat frame2, frameCopy;
                                        frame2 = frame;
                                        if( !frame2.empty() ) {
                                            if( frame->origin == IPL_ORIGIN_TL )
                                                frame2.copyTo( frameCopy );
                                            else
                                                flip( frame2, frameCopy, 0 );
                                            detectAndDraw(ctx, frameCopy);
                                            if (ctx->object_renderto) {
                                                imwrite(ctx->img_name_ptr, frameCopy);
                                                //MySaveImage(ctx->img_name_ptr, &frameCopy, true);
                                                gtk_image_set_from_file(GTK_IMAGE(ctx->object_renderto), ctx->img_name_ptr);
                                                unlink(ctx->img_name_ptr);
                                            }
                                        }
                                    }
                                    if (ctx->motion_detector) {
                                        update_mhi(ctx, frame);
                                        if ((ctx->secondary_renderto) && (ctx->motion)) {
                                            cvSaveImage(ctx->img_name_ptr, ctx->motion);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->secondary_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }
                            }
                        }
                        break;
                    case P_CAPTUREINIT:
                        break;
                    case P_RELEASECAPTURE:
                        if ((param.ToInt()) && (ctx->ctx)) {
                            cvReleaseCapture(&ctx->ctx);
                            ctx->ctx=0;
                        }
                        break;
                    case P_VIDEOFORMAT:
                        ctx->video_format=param;
                        break;
                    case P_VIDEOWRITERCREATE:
                        {
                            if ((param.ToInt()) && (ctx->ctx)) {
                                IplImage* frame = cvQueryFrame( ctx->ctx );
                                if (ctx->stream_data.Length()) {
                                    ctx->stream_data.LoadBuffer("",0);
                                }
                                if (ctx->video_writer) {
                                    cvReleaseVideoWriter(&ctx->video_writer);
                                    ctx->video_writer=0;
                                }
                                if (ctx->streamname.Length()) {
                                    unlink(ctx->streamname.c_str());
                                    MasterCtx->streamname=(char *)"";
                                }
                                if (frame) {
                                    CvVideoWriter *writer = 0;
                                    AnsiString fname((char *)"rec_");
                                    fname+=AnsiString((long)ctx->ctx);
                                    fname+=AnsiString(".avi");
                                    writer=cvCreateVideoWriter(fname.c_str(),CV_FOURCC(ctx->video_format[0],ctx->video_format[1],ctx->video_format[2],ctx->video_format[3]), ctx->fps,cvSize(frame->width,frame->height),1);
                                    ctx->video_writer=writer;
                                    if (writer)
                                        ctx->streamname=fname;
                                }
                            }
                        }
                        break;
                    case P_VIDEOWRITERDONE:
                        if ((param.ToInt()) && (ctx->video_writer)) {
                            cvReleaseVideoWriter(&ctx->video_writer);
                            ctx->video_writer=0;
                            if (ctx->streamname.Length()) {
                                ctx->stream_data.LoadFile(ctx->streamname.c_str());
                                unlink(ctx->streamname.c_str());
                                ctx->streamname=(char *)"";
                            }
                        }
                        break;
                    case P_CLEARSTREAMDATA:
                        if ((param.ToInt()) && (ctx->video_writer) && (ctx->stream_data.Length()))
                            ctx->stream_data.LoadBuffer("",0);
                        break;
                    case P_VIDEOUPDATEIMAGE:
                    case P_VIDEOUPDATE:
                        if ((ctx->ctx) && (ctx->video_writer)) {
                            IplImage* frame = cvQueryFrame( ctx->ctx );
                            if (!frame)
                                break;
                            cvWriteFrame(ctx->video_writer,frame);
                            if (prop_id==P_VIDEOUPDATEIMAGE) {
                                if (!ctx->img_name_ptr) {
                                    ctx->img_name=(char *)"cv_image";
                                    ctx->img_name+=AnsiString((long)ctx);
                                    ctx->img_name+=AnsiString("_");
                                    ctx->img_name+=AnsiString((long)ctx->renderto);
                                    ctx->img_name+=(char *)".";
                                    ctx->img_name+=ctx->format;
                                    ctx->img_name_ptr=ctx->img_name.c_str();
                                }
                                if (ctx->renderto) {
                                    cvSaveImage(ctx->img_name_ptr, frame);
                                    gtk_image_set_from_file(GTK_IMAGE(ctx->renderto), ctx->img_name_ptr);
                                    unlink(ctx->img_name_ptr);
                                }
                                if (ctx->tc.track_object) {
                                    track_object(ctx, frame);
                                    if (ctx->tc.image) {
                                        if (ctx->track_renderto) {
                                            cvSaveImage(ctx->img_name_ptr, ctx->tc.image);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->track_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }
                                if (ctx->cpair_len) {
                                    Mat frame2, frameCopy;
                                    frame2 = frame;
                                    if( !frame2.empty() ) {
                                        if( frame->origin == IPL_ORIGIN_TL )
                                            frame2.copyTo( frameCopy );
                                        else
                                            flip( frame2, frameCopy, 0 );
                                        detectAndDraw(ctx, frameCopy);
                                        if (ctx->object_renderto) {
                                            imwrite(ctx->img_name_ptr, frameCopy);
                                            //MySaveImage(ctx->img_name_ptr, &frameCopy, true);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->object_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }

                                if (ctx->motion_detector)
                                    update_mhi(ctx, frame);
                                if ((ctx->secondary_renderto) && (ctx->motion)) {
                                    cvSaveImage(ctx->img_name_ptr, ctx->motion);
                                    gtk_image_set_from_file(GTK_IMAGE(ctx->secondary_renderto), ctx->img_name_ptr);
                                    unlink(ctx->img_name_ptr);
                                }
                            }
                        }
                        break;
                    case P_VIDEOFPS:
                        ctx->fps=param.ToInt();
                        break;
                    case P_VIDEOINIT:
                        break;
                    case P_MOTIONDETECTOR:
                        ctx->motion_detector=(bool)param.ToInt();
                        break;
                    case P_MOTIONTRESHOLD:
                        ctx->motion_treshold=param.ToInt();
                        break;
                    case P_CLEAROBJECTS:
                        if (param.ToInt())
                            ctx->cpair_len=0;
                        break;
                    case P_SCALEOBJECTS:
                        ctx->scale=param.ToFloat();
                        break;
                    case P_TRACKVMIN:
                        ctx->tc.vmin=param.ToInt();
                        break;
                    case P_TRACKVMAX:
                        ctx->tc.vmax=param.ToInt();
                        break;
                    case P_TRACKSMIN:
                        ctx->tc.smin=param.ToInt();
                        break;
                    case P_TRACKOBJECT:
                        {
                            int params[4];
                            params[0]=params[1]=params[2]=params[3]=0;
                            char *s=param.c_str();
                            int len=param.Length();
                            int last=len-1;
                            int index=0;
                            AnsiString temp;
                            for (int i=0;i<len;i++) {
                                if ((s[i]==',') || (i==last)) {
                                    if (i==last)
                                        temp+=s[i];
                                    params[index]=temp.ToInt();
                                    index++;
                                    temp=(char *)"";
                                    if (index==4)
                                        break;
                                } else
                                    temp+=s[i];
                            }
                            ctx->tc.selection.x=MAX(params[0],0);
                            ctx->tc.selection.y=MAX(params[1],0);
                            ctx->tc.selection.width=params[2];
                            ctx->tc.selection.height=params[3];
                            //ctx->tc.select_object = 0;
                            if ( ctx->tc.selection.width > 0 && ctx->tc.selection.height > 0 )
                                ctx->tc.track_object = -1;
                        }
                        break;
                    case P_UPDATETRACK:
                        break;
                    case P_FRAMEWIDTH:
                        if (ctx->ctx)
                            cvSetCaptureProperty(ctx->ctx, CV_CAP_PROP_FRAME_WIDTH, param.ToInt());
                        break;
                    case P_FRAMEHEIGHT:
                        if (ctx->ctx)
                            cvSetCaptureProperty(ctx->ctx, CV_CAP_PROP_FRAME_HEIGHT,param.ToInt());
                        break;
                }
            }
            break;
       case INVOKE_GETPROPERTY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                int prop_id     =   va_arg(arg, int);
                char **str_buffer=   va_arg(arg, char **);
                int *buf_len     =   va_arg(arg, int*);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                buffered_text=(char *)"";
                CaptureContext *ctx=(CaptureContext *)context;
                switch (prop_id) {
                    case P_UPDATEFRAME:
                        if (ctx->ctx) {
                            IplImage* frame = 0;
                            frame = cvQueryFrame( ctx->ctx );
                            if (frame) {
                                if (!ctx->img_name_ptr) {
                                    ctx->img_name=(char *)"cv_image";
                                    ctx->img_name+=AnsiString((long)ctx);
                                    ctx->img_name+=AnsiString("_");
                                    ctx->img_name+=AnsiString((long)ctx->renderto);
                                    ctx->img_name+=(char *)".";
                                    ctx->img_name+=ctx->format;
                                    ctx->img_name_ptr=ctx->img_name.c_str();
                                }
                                cvSaveImage(ctx->img_name_ptr, frame);
                                //cvSaveImage(ctx->img_name_ptr, frame);

                                if (ctx->renderto)
                                    gtk_image_set_from_file(GTK_IMAGE(ctx->renderto), ctx->img_name_ptr);
                                buffered_text.LoadFile(ctx->img_name_ptr);
                                unlink(ctx->img_name_ptr);
                                if (ctx->tc.track_object) {
                                    track_object(ctx, frame);
                                    if (ctx->tc.image) {
                                        if (ctx->track_renderto) {
                                            cvSaveImage(ctx->img_name_ptr, ctx->tc.image);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->track_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }
                                if (ctx->cpair_len) {
                                    Mat frame2, frameCopy;
                                    frame2 = frame;
                                    if( !frame2.empty() ) {
                                        if( frame->origin == IPL_ORIGIN_TL )
                                            frame2.copyTo( frameCopy );
                                        else
                                            flip( frame2, frameCopy, 0 );
                                        detectAndDraw(ctx, frameCopy);
                                        if (ctx->object_renderto) {
                                            imwrite(ctx->img_name_ptr, frameCopy);
                                            //MySaveImage(ctx->img_name_ptr, &frameCopy, true);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->object_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }
                                if (ctx->motion_detector) {
                                    update_mhi(ctx, frame);
                                    if ((ctx->secondary_renderto) && (ctx->motion)) {
                                        cvSaveImage(ctx->img_name_ptr, ctx->motion);
                                        gtk_image_set_from_file(GTK_IMAGE(ctx->secondary_renderto), ctx->img_name_ptr);
                                        unlink(ctx->img_name_ptr);
                                    }
                                }
                            }
                        }
                        break;
                    case P_UPDATEMOTION:
                        if (ctx->ctx) {
                            IplImage* frame = 0;
                            frame = cvQueryFrame( ctx->ctx );
                            if (frame) {
                                if (!ctx->img_name_ptr) {
                                    ctx->img_name=(char *)"cv_image";
                                    ctx->img_name+=AnsiString((long)ctx);
                                    ctx->img_name+=AnsiString("_");
                                    ctx->img_name+=AnsiString((long)ctx->renderto);
                                    ctx->img_name+=(char *)".";
                                    ctx->img_name+=ctx->format;
                                    ctx->img_name_ptr=ctx->img_name.c_str();
                                }
                                if (ctx->renderto) {
                                    cvSaveImage(ctx->img_name_ptr, frame);
                                    gtk_image_set_from_file(GTK_IMAGE(ctx->renderto), ctx->img_name_ptr);
                                    unlink(ctx->img_name_ptr);
                                }
                                if (ctx->tc.track_object) {
                                    track_object(ctx, frame);
                                    if (ctx->tc.image) {
                                        if (ctx->track_renderto) {
                                            cvSaveImage(ctx->img_name_ptr, ctx->tc.image);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->track_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }
                                if (ctx->cpair_len) {
                                    Mat frame2, frameCopy;
                                    frame2 = frame;
                                    if( !frame2.empty() ) {
                                        if( frame->origin == IPL_ORIGIN_TL )
                                            frame2.copyTo( frameCopy );
                                        else
                                            flip( frame2, frameCopy, 0 );
                                        detectAndDraw(ctx, frameCopy);
                                        if (ctx->object_renderto) {
                                            imwrite(ctx->img_name_ptr, frameCopy);
                                            //MySaveImage(ctx->img_name_ptr, &frameCopy, true);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->object_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }
                                if (ctx->motion_detector) {
                                    update_mhi(ctx, frame);
                                    if (ctx->motion) {
                                        cvSaveImage(ctx->img_name_ptr, ctx->motion);
                                        if (ctx->secondary_renderto)
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->secondary_renderto), ctx->img_name_ptr);
                                        buffered_text.LoadFile(ctx->img_name_ptr);
                                        unlink(ctx->img_name_ptr);
                                    }
                                }
                            }
                        }
                        break;
                    case P_UPDATEOBJECT:
                        if (ctx->ctx) {
                            IplImage* frame = 0;
                            frame = cvQueryFrame( ctx->ctx );
                            if (frame) {
                                if (!ctx->img_name_ptr) {
                                    ctx->img_name=(char *)"cv_image";
                                    ctx->img_name+=AnsiString((long)ctx);
                                    ctx->img_name+=AnsiString("_");
                                    ctx->img_name+=AnsiString((long)ctx->renderto);
                                    ctx->img_name+=(char *)".";
                                    ctx->img_name+=ctx->format;
                                    ctx->img_name_ptr=ctx->img_name.c_str();
                                }
                                if (ctx->renderto) {
                                    cvSaveImage(ctx->img_name_ptr, frame);
                                    gtk_image_set_from_file(GTK_IMAGE(ctx->renderto), ctx->img_name_ptr);
                                    unlink(ctx->img_name_ptr);
                                }
                                if (ctx->tc.track_object) {
                                    track_object(ctx, frame);
                                    if (ctx->tc.image) {
                                        if (ctx->track_renderto) {
                                            cvSaveImage(ctx->img_name_ptr, ctx->tc.image);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->track_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }
                                if (ctx->cpair_len) {
                                    Mat frame2, frameCopy;
                                    frame2 = frame;
                                    if( !frame2.empty() ) {
                                        if( frame->origin == IPL_ORIGIN_TL )
                                            frame2.copyTo( frameCopy );
                                        else
                                            flip( frame2, frameCopy, 0 );
                                        detectAndDraw(ctx, frameCopy);
                                        imwrite(ctx->img_name_ptr, frameCopy);
                                        buffered_text.LoadFile(ctx->img_name_ptr);
                                        if (ctx->object_renderto)
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->object_renderto), ctx->img_name_ptr);
                                        unlink(ctx->img_name_ptr);
                                    }
                                }
                                if (ctx->motion_detector) {
                                    update_mhi(ctx, frame);
                                    if (ctx->motion) {
                                        if (ctx->secondary_renderto) {
                                            cvSaveImage(ctx->img_name_ptr, ctx->motion);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->secondary_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }
                            }
                        }
                        break;
                    case P_UPDATETRACK:
                        if (ctx->ctx) {
                            IplImage* frame = 0;
                            frame = cvQueryFrame( ctx->ctx );
                            if (frame) {
                                if (!ctx->img_name_ptr) {
                                    ctx->img_name=(char *)"cv_image";
                                    ctx->img_name+=AnsiString((long)ctx);
                                    ctx->img_name+=AnsiString("_");
                                    ctx->img_name+=AnsiString((long)ctx->renderto);
                                    ctx->img_name+=(char *)".";
                                    ctx->img_name+=ctx->format;
                                    ctx->img_name_ptr=ctx->img_name.c_str();
                                }
                                if (ctx->renderto) {
                                    cvSaveImage(ctx->img_name_ptr, frame);
                                    gtk_image_set_from_file(GTK_IMAGE(ctx->renderto), ctx->img_name_ptr);
                                    unlink(ctx->img_name_ptr);
                                }
                                if (ctx->tc.track_object) {
                                    track_object(ctx, frame);
                                    if (ctx->tc.image) {
                                        cvSaveImage(ctx->img_name_ptr, ctx->tc.image);
                                        buffered_text.LoadFile(ctx->img_name_ptr);
                                        if (ctx->track_renderto)
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->track_renderto), ctx->img_name_ptr);
                                        unlink(ctx->img_name_ptr);
                                    }
                                }
                                if (ctx->cpair_len) {
                                    Mat frame2, frameCopy;
                                    frame2 = frame;
                                    if( !frame2.empty() ) {
                                        if( frame->origin == IPL_ORIGIN_TL )
                                            frame2.copyTo( frameCopy );
                                        else
                                            flip( frame2, frameCopy, 0 );
                                        detectAndDraw(ctx, frameCopy);
                                        //buffered_text.LoadFile(ctx->img_name_ptr);
                                        if (ctx->object_renderto) {
                                            imwrite(ctx->img_name_ptr, frameCopy);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->object_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }
                                if (ctx->motion_detector) {
                                    update_mhi(ctx, frame);
                                    if (ctx->motion) {
                                        if (ctx->secondary_renderto) {
                                            cvSaveImage(ctx->img_name_ptr, ctx->motion);
                                            gtk_image_set_from_file(GTK_IMAGE(ctx->secondary_renderto), ctx->img_name_ptr);
                                            unlink(ctx->img_name_ptr);
                                        }
                                    }
                                }
                            }
                        }
                        break;
                    case P_CAPTUREINIT:
                        if (ctx->ctx)
                            buffered_text=(char *)"1";
                        else
                            buffered_text=(char *)"0";
                        break;
                    case P_VIDEOINIT:
                        if (ctx->video_writer)
                            buffered_text=(char *)"1";
                        else
                            buffered_text=(char *)"0";
                        break;
                    case P_STREAMDATA:
                        if (ctx->ctx) {
                            if (ctx->stream_data.Length())
                                buffered_text=ctx->stream_data;
                            else
                            if ((ctx->video_writer) && (ctx->streamname.Length())) {
                                buffered_text.LoadFile(ctx->streamname.c_str());
                            }
                        }
                        break;
                    case P_LASTHISTOGRAM:
                        if (ctx->tc.track_object) {
                            if (ctx->tc.histimg) {
                                if (!ctx->img_name_ptr) {
                                    ctx->img_name=(char *)"cv_image";
                                    ctx->img_name+=AnsiString((long)ctx);
                                    ctx->img_name+=AnsiString("_");
                                    ctx->img_name+=AnsiString((long)ctx->renderto);
                                    ctx->img_name+=(char *)".";
                                    ctx->img_name+=ctx->format;
                                    ctx->img_name_ptr=ctx->img_name.c_str();
                                }
                                cvSaveImage(ctx->img_name_ptr, ctx->tc.histimg);
                                buffered_text.LoadFile(ctx->img_name_ptr);
                                unlink(ctx->img_name_ptr);
                            }
                        }
                        break;
                }
                *str_buffer=buffered_text.c_str();
                *buf_len=buffered_text.Length();
                return E_NOERROR;
            }
            break;
        case INVOKE_SETSTREAM:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                char *target    =   va_arg(arg, char *);
                int len_target  =   va_arg(arg, int);
                char *value     =   va_arg(arg, char *);
                int len_value   =   va_arg(arg, int);
                va_end(arg);

                if ((!context) || (cls_id!=CUSTOM_CONTROL_ID))
                    break;

                String cascadeName;
                if (len_target)
                    cascadeName.assign(target);
                String nestedName;
                if (len_value)
                    nestedName.assign(value);
                buffered_text=(char *)"";
                CaptureContext *ctx=(CaptureContext *)context;
                if (ctx->cpair_len>=MAX_PAIRS)
                    break;
                if ((target) && (len_target)) {
                    if (!ctx->cpair[ctx->cpair_len].cascade.load(cascadeName)) {
                        fprintf(stderr, "ERROR: Could not load classifier cascade '%s'\n", target);
                    }
                }
                if ((value) && (len_value)) {
                    if (!ctx->cpair[ctx->cpair_len].nested.load(nestedName)) {
                        fprintf(stderr, "WARNING: Could not load classifier cascade '%s'\n", value); 
                    }
                }
                ctx->cpair_len++;
                return E_NOERROR;
            }
            break;
        case INVOKE_SETSECONDARYSTREAM:
            // not needed
            break;
        case INVOKE_CUSTOMMESSAGE:
            {
            }
            break;
        case INVOKE_CREATE:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void **handler  =   va_arg(arg, void **);
                int *pack_type  =   va_arg(arg, int *);
                va_end(arg);
                *pack_type=(int)2;

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                char *post_string=0;
                int len_post_string=0;
                RemoteInvoker(0,INVOKE_GETPOSTSTRING,&post_string,&len_post_string);
                int cam_id=0;
                
                AnsiString buf;
                if (len_post_string) {
                    buf.LoadBuffer(post_string, len_post_string);
                    cam_id=buf.ToInt();
                }

                CvCapture* capture = cvCaptureFromCAM(cam_id);
                if (!capture) {
                    return E_NOTIMPLEMENTED;
                }

                CaptureContext *ctx=new CaptureContext;
                ctx->ctx=capture;
                ctx->renderto=0;
                ctx->secondary_renderto=0;
                ctx->format=(char *)"png";
                ctx->video_format=(char *)"DIVX";
                ctx->video_writer=0;
                ctx->fps=25;
                ctx->buf=0;
                ctx->last=0;
                ctx->motion=0;
                ctx->motion_detector=false;
                ctx->motion_event=false;
                ctx->motion_treshold=30;
                ctx->storage=0;
                ctx->mhi=0;
                ctx->orient=0;
                ctx->mask=0;
                ctx->segmask=0;
                ctx->scale=0;
                ctx->cpair_len=0;
                ctx->detect_event=0;
                ctx->object_renderto=0;
                ctx->track_renderto=0;
                ctx->img_name_ptr=0;

                *handler=ctx;
                MasterCtx=ctx;
                return E_ACCEPTED_NOADD;
            }
            break;
        case INVOKE_DESTROY:
            {
                va_start(arg, INVOKE_TYPE);
                int cls_id      =   va_arg(arg, int);
                void *handler  =   va_arg(arg, void *);
                va_end(arg);

                if (cls_id!=CUSTOM_CONTROL_ID)
                    break;

                return E_ACCEPTED;
            }
            break;
        case INVOKE_INITLIBRARY:
            va_start(arg, INVOKE_TYPE);
            RemoteInvoker     =   va_arg(arg, INVOKER);
            //InitProcThread();
            va_end(arg);
            return E_NOERROR;
        case INVOKE_DONELIBRARY:
            if (MasterCtx) {
                if (MasterCtx->video_writer) {
                    cvReleaseVideoWriter(&MasterCtx->video_writer);
                    if (MasterCtx->streamname.Length())
                        unlink(MasterCtx->streamname.c_str());
                }
                MasterCtx->video_writer=0;
                if (MasterCtx->ctx)
                    cvReleaseCapture(&MasterCtx->ctx);
                MasterCtx->ctx=0;
            }
            return E_NOERROR;
        default:
            return E_NOTIMPLEMENTED;
    }
    return E_NOTIMPLEMENTED;
}


