#include "ConceptClient.h"
#include "mtrand.h"
#include "messages.h"
#include "callback.h"
#include "httpmessaging.h"

//#define DEBUG_SAY_WHERE std::cerr << ctime(0) << __FILE__ << ":" <<  __LINE__ << "\n";
#define DEBUG_SAY_WHERE

#define FAST_TCP

extern "C" {
    #include "rsa/rsa.h"
}
#include <time.h>

//CALLBACK_FUNC CConceptClient::CONCEPT_CALLBACK;
//Server SignalServer;
int CConceptClient::idle_called=0;
int CConceptClient::is_http=0;
int CConceptClient::PORT=DEFAULT_PORT;
AnsiString CConceptClient::HTTP_prefix;
AnsiString CConceptClient::HTTP_SendLink;
AnsiString CConceptClient::HTTP_RecvLink;
AnsiString CConceptClient::HTTP_PendingLink;

AnsiString CConceptClient::GenerateRandomAESKey(int len) {
    unsigned long init[4] = {time(NULL), time(NULL) >> 8, time(NULL) >> 16, time(NULL) >> 24}, length = 4;
    MTRand_int32 irand(init, length);
    char str[sizeof(long)+1];
    str[sizeof(long)]=0;
    AnsiString res;
    for (int i=0;i<len/4;i++) {
        *(long *)str=irand();
        for (int j=0;j<sizeof(long);j++)
            while (!str[j])
                str[j]=(char)irand();
        res+=str;
    }
    return res;
}

CConceptClient::CConceptClient(CALLBACK_FUNC cb,int secured, int debug, void *_notify) {
    this->CONCEPT_CALLBACK=cb;
    this->secured=secured;
    this->hThread=0;
    this->MSG_ENABLED=1;
    this->debug=debug;
    this->PARAMETERS=0;
    this->start_loop=false;
    this->is_up=false;
    this->lock1=false;
    this->lock2=false;
    this->notify=_notify;
    this->NON_CRITICAL_SECTION=0;
	idle_called=0;

#ifdef _WIN32
    WSADATA data;
    WSAStartup(MAKEWORD(1,1),&data);
#endif

    seminit(sem_message, 1);
    seminit(sem_done,  1);
    seminit(sem_send, 1);

    //LOCAL_PUBLIC_KEY.LoadFile(PUBLIC_KEY_FILE);
    //LOCAL_PRIVATE_KEY.LoadFile(PRIVATE_KEY_FILE);
}

int CConceptClient::ConnectHTML(char *HOST, int PORT) {
    AnsiString content="set servername ";
    content+=HOST;
    content+="\r\n";
    char *data=0;
    int len=0;
    AnsiString key;
    AnsiString prefix("http://");
    prefix+=HOST;
    if (PORT!=80) {
        prefix+=(char *)":";
        prefix+=AnsiString(PORT);
    }
    CConceptClient::PORT=PORT;
    prefix+=(char *)"/ConceptToHttp/";

    HTTPInit();

    HTTPGetRecv(prefix+"initsession", &data, &len);
    if ((data) && (len)) {
        key.LoadBuffer(data, len);
        HTTPFreeData(data);
    } else {
        std::cerr << "\nError getting session key from " << prefix.c_str() << "\n";
        return 0;
    }
    if (key.Length()!=32) {
        std::cerr << "\nInvalid session key length\n";
        return 0;
    }

    HTTP_prefix=prefix;
    HTTP_prefix+=key;

    HTTP_SendLink=HTTP_prefix+"/sendmessage";
    HTTP_RecvLink=HTTP_prefix+"/getmessage";
    HTTP_PendingLink=HTTP_prefix+"/havemessage";


    HTTPPostSend(HTTP_prefix+"/API", content.c_str(), content.Length());

    REMOTE_PUBLIC_KEY=0;
    LOCAL_PUBLIC_KEY=0;
    Called_HOST=HOST;

    return 1;
}

int CConceptClient::Connect(char *HOST, int PORT, long extra) {
    if (is_http)
        return this->ConnectHTML(HOST, PORT);

    struct sockaddr_in      sin;
    struct hostent          *hp;

    CConceptClient::PORT=PORT;
    if ((hp=gethostbyname(HOST))==0) {
        std::cerr << "\nUnable to resolve specified host\n";
        return 0;
    }

    memset(&sin,0,sizeof(sin));
                                  //INADDR_ANY;  
    sin.sin_addr.s_addr         = ((struct in_addr *)(hp->h_addr))->s_addr;
    sin.sin_family              = AF_INET;
    sin.sin_port                = htons(PORT);

    // creez socket-ul
    if ((CLIENT_SOCKET  =   socket(AF_INET, SOCK_STREAM, IPPROTO_TCP))==INVALID_SOCKET) {
        std::cerr << "\nError creating sockets\n";
        return 0;
    }

    int flag = 1;
#ifdef FAST_TCP
    setsockopt(                  CLIENT_SOCKET,   /* socket affected */
                                 IPPROTO_TCP,     /* set option at TCP level */
                                 TCP_NODELAY,     /* name of option */
                                 (char *) &flag,  /* the cast is historical cruft */
                                 sizeof(int));    /* length of option value */
#endif
    setsockopt(                  CLIENT_SOCKET,
                                 SOL_SOCKET,
                                 SO_KEEPALIVE,
                                 (char *) &flag,
                                 sizeof(int));

    if (connect(CLIENT_SOCKET,(struct sockaddr *)  &sin, sizeof(sin)) == INVALID_SOCKET) {
        std::cerr << "Error connecting to specified host\n";
        return 0;
    }
#ifdef PRINT_STATUS
    std::cerr << "\r[][][]:::::::::::: Generating random AES key ...                            \r";
#endif
    LOCAL_PUBLIC_KEY=GenerateRandomAESKey(RANDOM_KEY_SIZE);
    LOCAL_PRIVATE_KEY=LOCAL_PUBLIC_KEY;

    long cert_size=0;
    int received=0;
    int rec_tot=0;
    do {
        received=recv(CLIENT_SOCKET,(char *)&cert_size+rec_tot,sizeof(long)-rec_tot,0);
        rec_tot+=received;
    } while ((received>0) && (rec_tot<sizeof(long)));
    cert_size=ntohl(cert_size);
    
    if ((received<1) || (rec_tot!=sizeof(long))) {
        std::cerr << "\nConcept protocol breach\n";
        TRAY_NOTIFY("Concept protocol breach", "An invalid certificate size was received");
	    return 0;
    }

#ifdef PRINT_STATUS
    std::cerr << "\r[][][][]:::::::::: Receiving RSA certificate from server ...                \r";
#endif

    received=0;
    rec_tot=0;
    char buffer[4096];
    do {
        received=recv(CLIENT_SOCKET,buffer+rec_tot,cert_size-rec_tot,0);
        rec_tot+=received;
    } while ((received>0) && (rec_tot<cert_size));

    if ((received<1) || (rec_tot!=cert_size)) {
        std::cerr << "\nConcept protocol breach (" << cert_size << " bytes waited, received " << rec_tot << ")";
        TRAY_NOTIFY("Concept protocol breach", "Broken connection while receiving public certificate");
        return 0;
    }
    buffer[rec_tot]=0;
    REMOTE_PUBLIC_KEY=buffer;

#ifdef PRINT_STATUS
    std::cerr << "\r[][][][][]:::::::: Securing ...                                              \r";
#endif

    // criptez query-ul ...
    int in_content_size=LOCAL_PUBLIC_KEY.Length();
    int out_content_size=(((in_content_size / 48) + 1) * 64) + 256;
    char *out_content=new char[out_content_size+1];

    int encrypt_result = RSA_encrypt(   LOCAL_PUBLIC_KEY.c_str(), in_content_size,
	                                out_content, out_content_size,
                                        REMOTE_PUBLIC_KEY.c_str(), REMOTE_PUBLIC_KEY.Length());
    if (!encrypt_result) {
       delete[] out_content;
       std::cerr << "\nInvalid certificate\n";
       TRAY_NOTIFY("Invalid certificate", "An invalid certificate was received");
    }

#ifdef PRINT_STATUS
    std::cerr << "\r[][][][][][]:::::: Sending AES public key to server ...                      \r";
#endif

    long size=encrypt_result;
    long size_net=htonl(size);
    send(CLIENT_SOCKET,(char *)&size_net, sizeof(long), 0);
    send(CLIENT_SOCKET,out_content, size, 0);
    delete[] out_content;

    // setting hostname
    AnsiString content="set servername ";
    content+=HOST;
    content+="\r\n";

    send(CLIENT_SOCKET, content.c_str(), content.Length(), 0);

    REMOTE_PUBLIC_KEY=LOCAL_PUBLIC_KEY;

    Called_HOST=HOST;

    return 1;
}

THREAD_TYPE CConceptClient::ThreadMessage(LPVOID PARAM) {
    /*Parameters *P=(Parameters *)PARAM;
    send_message(P->Sender,P->ID,P->Target,P->Value,P->Owner->CLIENT_SOCKET,P->Owner->secured ? P->Owner->REMOTE_PUBLIC_KEY.c_str() : 0);
    delete P;*/
    CConceptClient *THIS_PTR=(CConceptClient *)PARAM;
    while (1) {
        semp(THIS_PTR->sem_send);
        int cnt=THIS_PTR->MessageStack.Count();
        semv(THIS_PTR->sem_send);
        if (cnt) {
                semp(THIS_PTR->sem_send);
                Parameters *P=(Parameters *)THIS_PTR->MessageStack.Remove(0);
                semv(THIS_PTR->sem_send);
                if (P) {
                    send_message(P->Sender,P->ID,P->Target,P->Value,P->Owner->CLIENT_SOCKET,P->Owner->secured ? P->Owner->REMOTE_PUBLIC_KEY.c_str() : 0, (PROGRESS_API)THIS_PTR->notify);
                    delete P;
                }
        }/* else
           Sleep(1);*/
        semv(THIS_PTR->sem_send);
        if (!cnt)
            Sleep(1);
    }
    return 0;
}

void CConceptClient::SendMessageNoWait(char *sender, int MSG, char *target, AnsiString value, char force) {
    if ((!MSG_ENABLED) && (!force))
        return;

    Parameters *P=new Parameters;
    P->Sender=sender;
    P->ID=MSG;
    P->Target=target;
    P->Value=value;
    P->Owner=this;
    semp(this->sem_send);
    this->MessageStack.Add(P,DATA_MESSAGE);
    semv(this->sem_send);
    /*DWORD TID;
    CreateThread(0,0,ThreadMessage,P,0,&TID);*/
}

int CConceptClient::WaitMessage(int MSG, Parameters *P) {
    P->ID=MSG;
    P->Owner=this;
	return wait_message(P, MSG, this->CLIENT_SOCKET,this->secured ? this->LOCAL_PRIVATE_KEY.c_str() : 0, (PROGRESS_API)this->notify);
}

long CConceptClient::SendMessage(char *sender, int MSG, char *target, AnsiString value, char force) {
    /*if ((!MSG_ENABLED) && (!force))
        return 0;
    return send_message(sender,MSG,target,value,CLIENT_SOCKET,secured ? REMOTE_PUBLIC_KEY.c_str() : 0);*/

    //SendMessageNoWait(sender, MSG, target, value, force);

    if ((!MSG_ENABLED) && (!force))
        return 0;

    Parameters *P=new Parameters;
    P->Sender=sender;
    P->ID=MSG;
    P->Target=target;
    P->Value=value;
    P->Owner=this;
    semp(this->sem_send);
    this->MessageStack.Add(P,DATA_MESSAGE);
    semv(this->sem_send);

    return 0;
}

long CConceptClient::SendBigMessage(char *sender, int MSG, char *target, AnsiString *value, char force) {
    if ((!MSG_ENABLED) && (!force))
        return 0;

    Parameters *P=new Parameters;
    P->Sender=sender;
    P->ID=MSG;
    P->Target=target;
    P->Value=*value;
    P->Owner=this;
    semp(this->sem_send);
    this->MessageStack.Add(P,DATA_MESSAGE);
    semv(this->sem_send);

    return 0;
}

long CConceptClient::SendFileMessage(char *sender, int MSG, char *target, char *filename, char force) {
    if ((!MSG_ENABLED) && (!force))
        return 0;

    Parameters *P=new Parameters;
    P->Sender=sender;
    P->ID=MSG;
    P->Target=target;
    P->Value=filename;
    P->Owner=this;
    semp(this->sem_send);
    this->MessageStack.Add(P,DATA_MESSAGE);
    semv(this->sem_send);
    return 0;
}

int CConceptClient::Disconnect() {
    if (CConceptClient::is_http) {
        HTTPPostSend(HTTP_prefix+"/API", "close\r\n", 7);
        return 0;
    } else {
        send(CLIENT_SOCKET,"close\r\n",7,0);
        return closesocket(CLIENT_SOCKET);
    }
}

INTEGER CConceptClient::NoErrors() {
    return 1;
}

int CConceptClient::Run(char *app_name, int pipe) {
    if (pipe>0) {
        AnsiString szpipe="pipe ";
        szpipe+=AnsiString((long)pipe);
        szpipe+="\r\n";
        if (CConceptClient::is_http)
            HTTPPostSend(HTTP_prefix+"/API", szpipe.c_str(), szpipe.Length());
        else
            send(CLIENT_SOCKET,szpipe.c_str(),szpipe.Length(),0);
    }

    AnsiString S="run ";
    if (!secured)
        S="runfast ";
    if (debug)
        S="rundebug ";
    S+=app_name;

    S+=(char *)"\r\n";
    if (CConceptClient::is_http)
        return HTTPPostSend(HTTP_prefix+"/API", S.c_str(), S.Length());
    else
        return send(CLIENT_SOCKET,S.c_str(),S.Length(),0);
}


THREAD_TYPE CConceptClient::ConceptMessageLoop(LPVOID LPSENDER) {
		//if (!Glib::thread_supported())
		//	Glib::thread_init();

        CConceptClient *SENDER=(CConceptClient *)LPSENDER;
        SENDER->is_up=true;
        static Parameters PARAM;
        PARAM.Owner=SENDER;
        // enable message sender ...
/*#ifdef _WIN32
        DWORD TID;
        CreateThread(0,0,ThreadMessage,SENDER,0,&TID);
#else
        //pthread *threadID;
        pthread_t threadID;
        pthread_create(&threadID, NULL, ThreadMessage,SENDER);
#endif*/
        g_thread_create(ThreadMessage, SENDER, FALSE, NULL);

        char clean=0;
        while (get_message(&PARAM,SENDER->CLIENT_SOCKET,SENDER->secured ? SENDER->LOCAL_PRIVATE_KEY.c_str() : 0, (PROGRESS_API)SENDER->notify)>0) {
            // Concept-message-loop
            SENDER->PARAMETERS=&PARAM;
            DEBUG_SAY_WHERE
            semp(SENDER->sem_message);//lock !
            DEBUG_SAY_WHERE
            semp(SENDER->sem_done); // lock !
            DEBUG_SAY_WHERE
            if (SENDER->DoCallback()) {
                clean=1;
                DEBUG_SAY_WHERE
                semv(SENDER->sem_message);
                DEBUG_SAY_WHERE
                break;
            }
            DEBUG_SAY_WHERE
            semv(SENDER->sem_message);
            DEBUG_SAY_WHERE
		}
        if (!clean) {
           SENDER->PARAMETERS=&PARAM;

           DEBUG_SAY_WHERE
           semp(SENDER->sem_message);
           DEBUG_SAY_WHERE
           semp(SENDER->sem_done);
           DEBUG_SAY_WHERE

           PARAM.ID=0x500;
           SENDER->DoCallback();

           DEBUG_SAY_WHERE
           semv(SENDER->sem_message);
           DEBUG_SAY_WHERE

           //CONCEPT_CALLBACK(&PARAM);
           //SENDER->SignalServer.message_notify(&PARAM);
        }
        return 0;
}

int CConceptClient::DoCallback() {
    CALL_RESULT=0;

    //if (start_loop) {
        semp(sem_done);
        semv(sem_done);

        /*while (PARAMETERS) {
            Sleep(1);
        }*/
    /*} else {
        Parameters *P=new Parameters;
        P->Sender   =   PARAMETERS->Sender;
        P->ID       =   PARAMETERS->ID;
        P->Target   =   PARAMETERS->Target;
        P->Value    =   PARAMETERS->Value;
        P->Owner    =   PARAMETERS->Owner;
        PARAMETERS=0;

        while (lock1);
        lock2=true;
        RMStack.Add(P,DATA_MESSAGE);
        lock2=false;
    }*/

    return CALL_RESULT;
}

void CConceptClient::WaitStartup() {
    while (!is_up)
        Sleep(10);
}

int CConceptClient::Iterate() {
    if (PARAMETERS) {
        DEBUG_SAY_WHERE
        CALL_RESULT=CONCEPT_CALLBACK(PARAMETERS);
        PARAMETERS=0;
        DEBUG_SAY_WHERE
        semv(sem_done);
        DEBUG_SAY_WHERE
        return CALL_RESULT;
    }
#ifdef _WIN32
	struct timeval s;
	s.tv_sec = 0;
	s.tv_usec = 250;
	select (0, NULL, NULL, NULL, &s);
#else
	usleep(250);
#endif
    return 0;
}

int CConceptClient::PreLoopMessage() {
    /*if (PARAMETERS) {
        on_idle_event();
        return 1;
    }*/
    while (RMStack.Count()) {
        lock1=true;
        while (lock2);
        Parameters *P=(Parameters *)RMStack.Remove(0);
        lock1=false;
        CALL_RESULT=CONCEPT_CALLBACK(P);
        delete P;

    }
    return 0;
}

bool CConceptClient::on_idle_event() {
    //semp(sem_done);
	if (idle_called) {
		return true;
	}
	idle_called=1;
    if (PARAMETERS) {
        CALL_RESULT=CONCEPT_CALLBACK(PARAMETERS);
        //---------------------------------------------------------------------------------------------//
        if ((NON_CRITICAL_SECTION) || ((PARAMETERS->ID>=MSG_CUSTOM_MESSAGE1) && (PARAMETERS->ID<=MSG_CUSTOM_MESSAGE16))) {
            int last_id=PARAMETERS->Sender.ToInt();
            while (have_messages(CLIENT_SOCKET)) {
				DEBUG_SAY_WHERE

                Parameters PARAM;
                PARAM.Owner=this;

                if (get_message(&PARAM,CLIENT_SOCKET,secured ? LOCAL_PRIVATE_KEY.c_str() : 0, (PROGRESS_API)notify, true)<=0)
                    break;

                CALL_RESULT=CONCEPT_CALLBACK(&PARAM);
                if (CALL_RESULT)
                    break;
                if (!NON_CRITICAL_SECTION) {
                    if (last_id!=PARAM.Sender.ToInt())
                        break;

                    if ((PARAM.ID<MSG_CUSTOM_MESSAGE1) || (PARAM.ID>MSG_CUSTOM_MESSAGE16))
                        break;
                }
            }
        }
        //---------------------------------------------------------------------------------------------//

        PARAMETERS=0;
        DEBUG_SAY_WHERE
        semv(sem_done);
        DEBUG_SAY_WHERE
    } else {
        LocalIDLE();
        Sleep(1);
    }
	idle_called=0;
    return true;
}

void CConceptClient::StartMessaging() {
    //SignalServer.signal_message_notify().connect( sigc::mem_fun(*this, &CConceptClient::DoCallback) );
/*#ifdef _WIN32
    DWORD TID;
    hThread=CreateThread(0,0,(LPTHREAD_START_ROUTINE)ConceptMessageLoop,this,0,&TID);
#else
    pthread_t threadID;
    pthread_create(&threadID, NULL, ConceptMessageLoop,this);
#endif*/
    g_thread_create(ConceptMessageLoop, this, FALSE, NULL);
}

void CConceptClient::NotifyStartLoop() {
    start_loop=true;
    PreLoopMessage();
}

int CConceptClient::Dump(char * dump) {
    /*std::ofstream out;
    out.open(OUTPUT_DUMP,  std::ios_base::out | std::ios_base::app);
    out << dump;
    out.close();*/
    std::cout << dump;
    return 0;
}

int CConceptClient::EnableMessaging(int i) {
    this->MSG_ENABLED=(i!=0);
    return 0;
}
void CConceptClient::SetCritical(int status) {
    NON_CRITICAL_SECTION=status;
}

CConceptClient::~CConceptClient(void) {
#ifdef _WIN32
    WSACleanup();
#endif
}
