//#pragma once
#ifndef __CONCEPT_CLIENT_H
#define __CONCEPT_CLIENT_H
// uncomment to print to stdin (to enable console)
//#define PRINT_STATUS

#ifdef _WIN32
    #include <windows.h>
    #include <winsock.h>
    #define THREAD_TYPE2 DWORD WINAPI
#else
    #define SOCKET int
    #define INVALID_SOCKET -1
    #define SOCKET_ERROR -1
    #define DWORD long
    #define LPVOID void*
    #define WINAPI
    #define THREAD_TYPE2 LPVOID
    #define closesocket close

    #include <sys/types.h>
    #include <sys/socket.h>
    #include <sys/wait.h>
    #include <netinet/in.h>
    #include <netdb.h>
    #include <dlfcn.h>   // dlopen
    #include <unistd.h>
    #include <pthread.h>

    #include <sys/select.h>
    #include <netinet/tcp.h>
    #include <arpa/inet.h>

    #define HANDLE 	void *
    #define HMODULE     HANDLE

    #define Sleep(x) usleep(x*1000)
#endif


#include "AnsiString.h"
#include "AnsiList.h"
#include <iostream>
#include <fstream>
//#include "comunicator.h"
#include "semhh.h"

#include <glib.h>

#define THREAD_TYPE gpointer

#define WM_USER_CREATE_CONTROL      WM_USER+0x100

#define DEFAULT_PORT        2662
#define DEFAULT_HTTP_PORT   80

#define OUTPUT_DUMP         "core-output-dump.txt"

#define     T_NUMBER     0x02
#define     T_STRING     0x03

#define INTEGER             int

//#define PUBLIC_KEY_FILE       "client.certificate"
//#define PRIVATE_KEY_FILE	"client.key"

#define AES_KEY_FILE		"aes.key"
#define PUBLIC_KEY_FILE		AES_KEY_FILE
#define PRIVATE_KEY_FILE	AES_KEY_FILE
#define RANDOM_KEY_SIZE     16

#define CONCEPT_DLL_LINK        INTEGER
#define CONCEPT_DLL_RESULT      void *
#define SYSTEM_SOCKET           SOCKET

extern "C" {
#ifdef _WIN32
	#define CONCLIENT_DLL_API         __declspec(dllexport) int _cdecl
#else
	#define CONCLIENT_DLL_API         int
#endif

CONCLIENT_DLL_API ConceptClientRun(void *handle, const char *host, const char *app, int secured, void *reserved);
}

class CConceptClient;

typedef struct Parameters {
    AnsiString Sender;
    int ID;
    AnsiString Target;
    AnsiString Value;
    CConceptClient *Owner;
} TParameters;    

typedef int (*CALLBACK_FUNC)(Parameters *PARAM);

class CConceptClient {
    private:
        HHSEM  sem_message;
        HHSEM  sem_send;
        HHSEM  sem_done;

        static THREAD_TYPE         ThreadMessage(LPVOID PARAM);

        AnsiString                  LOCAL_PUBLIC_KEY;
        AnsiString                  LOCAL_PRIVATE_KEY;
        AnsiString                  REMOTE_PUBLIC_KEY;
        int secured;
        int debug;

        int                         waitconfirmation;
        HANDLE                      hThread;
        
        int         MSG_ENABLED;
        int         NON_CRITICAL_SECTION;


        AnsiList    MessageStack;
        AnsiList    RMStack;
        CALLBACK_FUNC CONCEPT_CALLBACK;

        Parameters *PARAMETERS;
        int CALL_RESULT;
        /*Server SignalServer;*/
        virtual int DoCallback();
        bool start_loop;
        bool is_up;
        bool lock1;
        bool lock2;
        bool more_data;

        void *notify;
		static int idle_called;
    public:
        static int PORT;
        static int is_http;
        static AnsiString HTTP_prefix;
        static AnsiString HTTP_SendLink;
        static AnsiString HTTP_RecvLink;
        static AnsiString HTTP_PendingLink;

        AnsiString Called_HOST;

        bool        on_idle_event();
        void        NotifyStartLoop();
        int         PreLoopMessage();
        int         Iterate();
        void        WaitStartup();
        void        SetCritical(int status);

        SOCKET      CLIENT_SOCKET;
        void        StartMessaging();
        int         Connect(char *HOST, int PORT=DEFAULT_PORT, long extra=0);
        int         ConnectHTML(char *HOST, int PORT=DEFAULT_HTTP_PORT);
        int         Disconnect();
        int         Run(char *app_name, int pipe=-1);
        int         EnableMessaging(int);

        int			WaitMessage(int MSG, Parameters *P);
        long        SendMessage(char *sender, int MSG, char *target, AnsiString value, char force=0);
        long        SendBigMessage(char *sender, int MSG, char *target, AnsiString *value, char force=0);
        long        SendFileMessage(char *sender, int MSG, char *target, char *filename, char force=0);
        void        SendMessageNoWait(char *sender, int MSG, char *target, AnsiString value, char force=0);


        AnsiString  GenerateRandomAESKey(int);

        static THREAD_TYPE ConceptMessageLoop(LPVOID);

        int         NoErrors();
        int         Dump(char *);
        CConceptClient(CALLBACK_FUNC cb, int secured, int debug, void *_notify=0);
        virtual ~CConceptClient();
};

#endif // __CONCEPT_CLIENT_H
