#include "PropertiesPanel.h"
#include "BasicMessages.h"
#include <gdk/gdkkeysyms.h>

PropertiesPanel::PropertiesPanel() {
    handle=GTK_VBOX(gtk_vbox_new(FALSE, 0));
    pTable=0;
    CC=0;
    font=pango_font_description_from_string("bold");
}

void PropertiesPanel::MapOnChangeEvent(CConceptClient *_CC, int _ID) {
    CC=_CC;
    ID=AnsiString((long)_ID);
}

PropertiesPanel::~PropertiesPanel() {
}

void PropertiesPanel::changed(bool recursive) {
    int len=Properties.Count();
    int temp=-1;
    int force=0;
    for (int i=0;i<len;i++) {
        LocalProperty *LP=(LocalProperty *)Properties.Item(i);
        AnsiString value((char *)"");
        switch (LP->type) {
                case PROP_INSTANT:
                        value=(char *)gtk_entry_get_text((GtkEntry *)LP->renderer);
                        break;
                case PROP_OPTION:
                        value=(char *)gtk_combo_box_get_active_text((GtkComboBox *)LP->renderer);
                        break;
                case PROP_OPTIONEDIT:
                        if (!recursive) {
                            temp=gtk_combo_box_get_active((GtkComboBox *)LP->renderer);
                            if (temp<0)
                                continue;
                        }
                        
                        value=(char *)gtk_entry_get_text(GTK_ENTRY(gtk_bin_get_child(GTK_BIN(LP->renderer))));
                        break;
                case PROP_COLOR:
                        {
                            GdkColor color;
                            gtk_color_button_get_color(GTK_COLOR_BUTTON(LP->renderer),&color);
                            long l_color=(color.red/0xFF)*0x10000;
                            l_color+=(color.green/0xFF)*0x100;
                            l_color+=color.blue/0xFF;
                            if (l_color>0xFFFFFF)
                                l_color=0xFFFFFF;
                            value=(long)l_color;
                        }
                        break;
                case PROP_FONT:
                        value=(char *)gtk_font_button_get_font_name(GTK_FONT_BUTTON(LP->renderer));
                        break;
                case PROP_FILE:
                        if (!gtk_widget_get_sensitive(GTK_WIDGET(LP->renderer))) {
                            gtk_widget_set_sensitive(GTK_WIDGET(LP->renderer),TRUE);
                            force=1;
                        }
                        break;
                case PROP_STATIC:
                        break;
        }
        if (((LP->value!=value) && (LP->type!=PROP_FILE)) || (force)) {
            LP->changed=1;

            LP->value=value;
            // notify the client !!!
            if (CC) {
                AnsiString param=LP->name;
                param+=(char *)"=";
                param+=value;
                CC->SendMessage(ID.c_str(),MSG_EVENT_FIRED,AnsiString((long)EVENT_ON_CHANGED),param.c_str());
            }
            //break;
        }
    }
}

gboolean PropertiesPanel::changed_cbe(GtkComboBox *widget, gpointer user_data) {
    ((PropertiesPanel *)user_data)->changed(false);
    return FALSE;
}

gboolean PropertiesPanel::focus_out_event(GtkWidget *widget, GdkEventFocus *event, gpointer user_data) {
    ((PropertiesPanel *)user_data)->changed(true);
    return FALSE;
}

gboolean PropertiesPanel::key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data) {
    if (event->keyval == GDK_Return) { 
        ((PropertiesPanel *)user_data)->changed(true);
    }
    return FALSE;
}

gboolean PropertiesPanel::custom_click(GtkWidget *widget, gpointer user_data) {
    gtk_widget_set_sensitive(widget, FALSE);
    ((PropertiesPanel *)user_data)->changed(true);
    return FALSE;
}

gboolean PropertiesPanel::color_set(GtkWidget *widget, gpointer user_data) {
    ((PropertiesPanel *)user_data)->changed(true);
    return FALSE;
}

gboolean PropertiesPanel::font_set(GtkWidget *widget, gpointer user_data) {
    ((PropertiesPanel *)user_data)->changed(true);
    return FALSE;
}

void PropertiesPanel::SetProperties(AnsiString prop_xml) {
    XMLNode xMainNode=XMLNode::parseString(prop_xml.c_str(),"<pr>");
    int prop_count=xMainNode.nChildNode("pp");
    int pindex;
    Properties.Clear();
    if (pTable) {
        gtk_container_remove(GTK_CONTAINER(handle), GTK_WIDGET(pTable));
        // managed by Gtk ? :)
    }
    pTable=GTK_TABLE(gtk_table_new(1,1,FALSE));

    gtk_box_pack_start(GTK_BOX(handle), GTK_WIDGET(pTable), FALSE, FALSE, 0);
    gtk_widget_show(GTK_WIDGET(pTable));
    bool important;

    for (int j=0;j<prop_count;j++) {
        gtk_table_resize(pTable,j+1,2);
        pindex=j;
        XMLNode xPNode=xMainNode.getChildNode("pp",&pindex);

        AnsiString PropName=(char *)xPNode.getAttribute("n");
        char Type=AnsiString((char *)xPNode.getAttribute("t")).ToInt();
        AnsiString Value=(char *)xPNode.getAttribute("v");

        important=false;
        if (Type<0) {
            Type*=-1;
            important=true;
        }

        // build data line
        GtkLabel *label=GTK_LABEL(gtk_label_new((AnsiString(" ")+PropName+" ").c_str()));
        gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);

        if (important)
            gtk_widget_modify_font(GTK_WIDGET(label), font);

        gtk_table_attach(GTK_TABLE(pTable), GTK_WIDGET(label), 0, 1, j, j+1, (GtkAttachOptions)(GTK_FILL | GTK_SHRINK), (GtkAttachOptions)GTK_SHRINK, 0, 0);

        //add(*label);

        gtk_widget_show(GTK_WIDGET(label));

        LocalProperty *LP=new LocalProperty;
        Properties.Add(LP,DATA_PROPERTY);

        LP->name=PropName;
        LP->value=Value;
        LP->type=Type;
        LP->changed=0;

        if (Type==PROP_OPTION) {
            int opt_count=xPNode.nChildNode("o");
            int oindex;
            GtkComboBox *combo=GTK_COMBO_BOX(gtk_combo_box_new_text());
            //GtkListStore *model= gtk_list_store_new (1, G_TYPE_STRING);

            int index=0;

            for (int k=0;k<opt_count;k++) {
                oindex=k;
                XMLNode xONode=xPNode.getChildNode("o",&oindex);
                AnsiString option=(char *)xONode.getText();
                // add option !
                if (option==Value)
                    index=k;

                gtk_combo_box_append_text(combo, option.c_str());
            }
            gtk_table_attach(GTK_TABLE(pTable), GTK_WIDGET(combo), 1, 2, j, j+1, (GtkAttachOptions)(GTK_FILL | GTK_EXPAND), (GtkAttachOptions)GTK_SHRINK, 0, 0);
            gtk_combo_box_set_active(combo, index);
            gtk_widget_show(GTK_WIDGET(combo));

            //combo->signal_changed().connect(sigc::mem_fun(*this, &PropertiesPanel::changed_cbe));
            g_signal_connect(G_OBJECT(combo), "changed", G_CALLBACK(PropertiesPanel::changed_cbe), this);

            gtk_widget_set_size_request(GTK_WIDGET(combo), 0, -1);

            LP->renderer=GTK_WIDGET(combo);
        } else
        if (Type==PROP_OPTIONEDIT) {
            int opt_count=xPNode.nChildNode("o");
            int oindex;
            GtkComboBoxEntry *combo=GTK_COMBO_BOX_ENTRY(gtk_combo_box_entry_new_text());
            gtk_table_attach(GTK_TABLE(pTable), GTK_WIDGET(combo), 1, 2, j, j+1, (GtkAttachOptions)(GTK_FILL | GTK_EXPAND), (GtkAttachOptions)GTK_SHRINK, 0, 0);


            for (int k=0;k<opt_count;k++) {
                oindex=k;
                XMLNode xONode=xPNode.getChildNode("o",&oindex);
                AnsiString option=(char *)xONode.getText();
                // add option !
                gtk_combo_box_append_text(GTK_COMBO_BOX(combo), option.c_str());
            }
            gtk_entry_set_text(/*GTK_ENTRY(gtk_bin_get_child(GTK_BIN(LP->renderer)))*/GTK_ENTRY(gtk_bin_get_child(GTK_BIN(combo))),Value.c_str());
            g_signal_connect(/*G_OBJECT(gtk_bin_get_child(GTK_BIN(LP->renderer)))*/gtk_bin_get_child(GTK_BIN(combo)), "key-press-event", G_CALLBACK(PropertiesPanel::key_press_event), this);
            gtk_widget_show(GTK_WIDGET(combo));
            g_signal_connect(G_OBJECT(combo), "changed", G_CALLBACK(PropertiesPanel::changed_cbe), this);
            g_signal_connect(/*G_OBJECT(gtk_bin_get_child(GTK_BIN(LP->renderer)))*/gtk_bin_get_child(GTK_BIN(combo)), "focus-out-event", G_CALLBACK(PropertiesPanel::focus_out_event), this);
            LP->renderer=GTK_WIDGET(combo);

            gtk_widget_set_size_request(GTK_WIDGET(combo), 0, -1);
        } else
        if (Type==PROP_COLOR) {
            int color=Value.ToInt();
            int red=color/0x10000;
            color%=0x10000;
            int green=color/0x100;
            color%=0x100;
            int blue=color;
            GdkColor uset_color;
            uset_color.red=((unsigned short)red*0xFF);
            uset_color.green=((unsigned short)green*0xFF);
            uset_color.blue=((unsigned short)blue*0xFF);

            GtkWidget *button=gtk_color_button_new_with_color(&uset_color);
            gtk_table_attach(GTK_TABLE(pTable), GTK_WIDGET(button), 1, 2, j, j+1, (GtkAttachOptions)(GTK_FILL | GTK_EXPAND), (GtkAttachOptions)GTK_SHRINK, 0, 0);            
            gtk_widget_show(GTK_WIDGET(button));
            g_signal_connect(G_OBJECT(button), "color-set", G_CALLBACK(PropertiesPanel::font_set), this);

            LP->renderer=GTK_WIDGET(button);
        } else
        if (Type==PROP_FONT) {
            GtkWidget *button=0;
            if (Value.Length()) {
                button=gtk_font_button_new_with_font(Value.c_str());
            } else {
                button=gtk_font_button_new();
            }
            gtk_table_attach(GTK_TABLE(pTable), GTK_WIDGET(button), 1, 2, j, j+1, (GtkAttachOptions)(GTK_FILL | GTK_EXPAND), (GtkAttachOptions)GTK_SHRINK, 0, 0);            
            gtk_widget_show(GTK_WIDGET(button));
            g_signal_connect(G_OBJECT(button), "font-set", G_CALLBACK(PropertiesPanel::font_set), this);

            LP->renderer=GTK_WIDGET(button);
        } else
        if (Type==PROP_FILE) {
            GtkWidget *button=0;
            if (Value.Length()) {
                button=gtk_button_new_with_label(Value.c_str());
            } else {
                button=gtk_button_new();
            }
            gtk_table_attach(GTK_TABLE(pTable), GTK_WIDGET(button), 1, 2, j, j+1, (GtkAttachOptions)(GTK_FILL | GTK_EXPAND), (GtkAttachOptions)GTK_SHRINK, 0, 0);            
            gtk_widget_show(GTK_WIDGET(button));
            g_signal_connect(G_OBJECT(button), "clicked", G_CALLBACK(PropertiesPanel::custom_click), this);

            LP->renderer=GTK_WIDGET(button);
        } else
        if (Type==PROP_STATIC) {
            GtkEntry *entry=GTK_ENTRY(gtk_entry_new());
            gtk_table_attach(GTK_TABLE(pTable), GTK_WIDGET(entry), 1, 2, j, j+1, (GtkAttachOptions)(GTK_FILL | GTK_EXPAND), (GtkAttachOptions)GTK_SHRINK, 0, 0);
            gtk_entry_set_text(entry, Value.c_str());
            gtk_entry_set_editable(entry, FALSE);
            gtk_widget_show(GTK_WIDGET(entry));
            LP->renderer=GTK_WIDGET(entry);
            gtk_widget_set_size_request(GTK_WIDGET(entry), 0, -1);
            Properties.Delete(Properties.Count()-1);
        } else {
            GtkEntry *entry=GTK_ENTRY(gtk_entry_new());
            gtk_table_attach(GTK_TABLE(pTable), GTK_WIDGET(entry), 1, 2, j, j+1, (GtkAttachOptions)(GTK_FILL | GTK_EXPAND), (GtkAttachOptions)GTK_SHRINK, 0, 0);

            //add(*entry);
            gtk_entry_set_text(entry, Value.c_str());
            gtk_widget_show(GTK_WIDGET(entry));
            g_signal_connect(G_OBJECT(entry), "key-press-event", G_CALLBACK(PropertiesPanel::key_press_event), this);
            g_signal_connect(G_OBJECT(entry), "focus-out-event", G_CALLBACK(PropertiesPanel::focus_out_event), this);
            LP->renderer=GTK_WIDGET(entry);
            gtk_widget_set_size_request(GTK_WIDGET(entry), 0, -1);
        }
    }
    changed(false);
}

AnsiString PropertiesPanel::GetPropertiesValues() {
}

AnsiString PropertiesPanel::GetChangedPropertiesValues(){
}

